import time
from typing import Callable, List, Optional, Union

from PIL import Image
from RPA.Desktop.keywords import (
    LibraryContext,
    keyword,
    screen,
    ElementNotFound,
    MultipleElementsFound,
    TimeoutException,
    HAS_RECOGNITION,
)

from RPA.core.geometry import Point, Region, Undefined
from RPA.core.locators import (
    Locator,
    LocatorType,
    PointLocator,
    OffsetLocator,
    RegionLocator,
    SizeLocator,
    ImageLocator,
    OcrLocator,
    syntax,
)

if HAS_RECOGNITION:
    from RPA.recognition import templates, ocr  # pylint: disable=no-name-in-module

Geometry = Union[Point, Region, Undefined]


def ensure_recognition():
    if not HAS_RECOGNITION:
        raise ValueError(
            "Locator type not supported, please install the "
            "rpaframework-recognition package"
        )


def transform(
    regions: List[Region], source: Region, destination: Region
) -> List[Region]:
    """Transform given regions from a local coordinate system to a
    global coordinate system.

    Takes into account location and scaling of the regions.
    Assumes that the aspect ratio does not change.

    :param regions: List of regions to transform
    :param source: Dimensions of local coordinate system
    :param destination: Position/scale of local coordinates in the global scope
    """
    scale = float(destination.height) / float(source.height)

    transformed = []
    for region in regions:
        region = region.scale(scale)
        region = region.move(destination.left, destination.top)
        transformed.append(region)

    return transformed


def clamp(minimum: float, value: float, maximum: float) -> float:
    """Clamp value between given minimum and maximum."""
    return max(minimum, min(value, maximum))


class FinderKeywords(LibraryContext):
    """Keywords for locating elements."""

    def __init__(self, ctx):
        super().__init__(ctx)
        self._resolver = syntax.Resolver(self._find, self.locators_path)

        self.timeout = 3.0
        if HAS_RECOGNITION:
            self.confidence = templates.DEFAULT_CONFIDENCE
        else:
            self.confidence = 80.0

    def _find(self, base: Geometry, locator: LocatorType) -> List:
        """Internal method for resolving and searching locators."""
        if isinstance(locator, (Region, Point)):
            return [locator]

        self.logger.debug("Finding locator: %s", locator)

        finders = {
            PointLocator: self._find_point,
            OffsetLocator: self._find_offset,
            RegionLocator: self._find_region,
            SizeLocator: self._find_size,
            ImageLocator: self._find_templates,
            OcrLocator: self._find_ocr,
        }

        for klass, finder in finders.items():
            if isinstance(locator, klass):
                return finder(base, locator)

        raise NotImplementedError(f"Unsupported locator: {locator}")

    def _find_point(self, base: Geometry, point: PointLocator):
        """Find absolute point on screen. CAn not be based on existing value."""
        if not isinstance(base, Undefined):
            self.logger.warning("Using absolute point coordinates")

        result = Point(point.x, point.y)
        return [result]

    def _find_offset(self, base: Geometry, offset: OffsetLocator):
        """Find pixel offset from given base value, or if no base,
        offset from current mouse position.
        """
        if isinstance(base, Undefined):
            position = self.ctx.get_mouse_position()
        elif isinstance(base, Region):
            position = base.center
        else:
            position = base

        result = position.move(offset.x, offset.y)
        return [result]

    def _find_region(self, base: Geometry, region: RegionLocator):
        """Find absolute region on screen. Can not be based on existing value."""
        if not isinstance(base, Undefined):
            self.logger.warning("Using absolute region coordinates")

        position = Region(region.left, region.top, region.right, region.bottom)
        return [position]

    def _find_size(self, base: Geometry, size: SizeLocator):
        """Find region of fixed size around base, or origin if no base defined."""
        if isinstance(base, Undefined):
            return Region.from_size(0, 0, size.width, size.height)

        if isinstance(base, Region):
            center = base.center
        else:
            center = base

        left = center.x - size.width // 2
        top = center.y - size.height // 2

        result = Region.from_size(left, top, size.width, size.height)
        return [result]

    def _find_templates(self, base: Geometry, locator: ImageLocator) -> List[Region]:
        """Find all regions that match given image template,
        inside the combined virtual display.
        """
        ensure_recognition()

        if isinstance(base, Undefined):
            region = None
        elif isinstance(base, Region):
            region = base
        else:
            raise ValueError(f"Unsupported search specifier for template: {base}")

        confidence = locator.confidence or self.confidence
        self.logger.info(
            "Searching for image '%s' (region: %s, confidence: %.1f)",
            locator.path,
            region or "display",
            confidence,
        )

        def finder(image: Image.Image) -> List[Region]:
            try:
                return templates.find(
                    image=image,
                    template=locator.path,
                    confidence=confidence,
                    region=region,
                )

            except templates.ImageNotFoundError:
                return []

        return self._find_from_displays(finder)

    def _find_ocr(self, base: Geometry, locator: OcrLocator) -> List[Region]:
        """Find the position of all blocks of text that match the given string,
        inside the combined virtual display.
        """
        ensure_recognition()

        if isinstance(base, Undefined):
            region = None
        elif isinstance(base, Region):
            region = base
        else:
            raise ValueError(f"Unsupported search specifier for OCR: {base}")

        confidence = locator.confidence or self.confidence
        self.logger.info(
            "Searching for text '%s' (region: %s, confidence: %.1f)",
            locator.text,
            region or "display",
            confidence,
        )

        def finder(image: Image.Image) -> List[Region]:
            matches = ocr.find(
                image=image,
                text=locator.text,
                confidence=confidence,
                region=region,
            )

            return [match["region"] for match in matches]

        return self._find_from_displays(finder)

    def _find_from_displays(
        self, finder: Callable[[Image.Image], List[Region]]
    ) -> List[Region]:
        """Call finder function for each display and return
        a list of found regions.

        :param finder: Callable that searches an image
        """
        matches = []
        screenshots = []

        # Search all displays, and map results to combined virtual display

        start_time = time.time()
        for display in screen.displays():
            image = screen.grab(display)
            regions = finder(image)

            for region in regions:
                region = region.resize(5)
                screenshot = image.crop(region.as_tuple())
                screenshots.append(screenshot)

            local = Region.from_size(0, 0, image.size[0], image.size[1])
            regions = transform(regions, local, display)
            matches.extend(regions)

        # Log matches and preview images

        duration = time.time() - start_time
        plural = "es" if len(matches) != 1 else ""

        self.logger.debug("Searched in %.2f seconds", duration)
        self.logger.info("Found %d match%s", len(matches), plural)

        for match, screenshot in zip(matches, screenshots):
            screen.log_image(screenshot, size=400)
            self.logger.info(match)

        return matches

    @keyword
    def find_elements(self, locator: LocatorType) -> List[Geometry]:
        """Find all elements defined by locator, and return their positions.

        :param locator: Locator string

        Example:

        .. code-block:: robotframework

            ${matches}=    Find elements    image:icon.png
            FOR    ${match}  IN  @{matches}
                Log    Found icon at ${match.right}, ${match.top}
            END
        """
        self.logger.info("Resolving locator: %s", locator)

        if isinstance(locator, (Locator, Region, Point)):
            return self._find(Undefined(), locator)
        else:
            return self._resolver.dispatch(str(locator))

    @keyword
    def find_element(self, locator: LocatorType) -> Geometry:
        """Find an element defined by locator, and return its position.
        Raises ``ElementNotFound`` if` no matches were found, or
        ``MultipleElementsFound`` if there were multiple matches.

        :param locator: Locator string

        Example:

        .. code-block:: robotframework

            ${match}=    Find element    image:logo.png
            Log    Found logo at ${match.right}, ${match.top}
        """
        matches = self.find_elements(locator)

        if not matches:
            raise ElementNotFound(f"No matches found for: {locator}")

        if len(matches) > 1:
            # TODO: Add run-on-error support and maybe screenshotting matches?
            raise MultipleElementsFound(
                "Found {count} matches for: {locator} at locations {matches}".format(
                    count=len(matches), locator=locator, matches=matches
                )
            )

        return matches[0]

    @keyword
    def wait_for_element(
        self,
        locator: LocatorType,
        timeout: Optional[float] = None,
        interval: float = 0.5,
    ) -> Geometry:
        """Wait for an element defined by locator to exist, or
        raise a TimeoutException if none were found within timeout.

        :param locator: Locator string

        Example:

        .. code-block:: robotframework

            Wait for element    alias:CookieConsent    timeout=30
            Click    image:%{ROBOT_ROOT}/accept.png
        """
        if timeout is None:
            timeout = self.timeout

        interval = float(interval)
        end_time = time.time() + float(timeout)

        error = "Operation timed out"
        while time.time() <= end_time:
            start = time.time()
            try:
                return self.find_element(locator)
            except (ElementNotFound, MultipleElementsFound) as err:
                error = err

            duration = time.time() - start
            if duration < interval:
                time.sleep(interval - duration)

        raise TimeoutException(error)

    @keyword
    def set_default_timeout(self, timeout: float = 3.0):
        """Set the default time to wait for elements.

        :param timeout: Time in seconds
        """
        self.timeout = max(0.0, float(timeout))

    @keyword
    def set_default_confidence(self, confidence: float = None):
        """Set the default template matching confidence.

        :param confidence: Value from 1 to 100
        """
        if confidence is None:
            confidence = templates.DEFAULT_CONFIDENCE if HAS_RECOGNITION else 80.0

        self.confidence = clamp(1.0, float(confidence), 100.0)



import pytest
from pathlib import Path
from RPA.Images import Images, TemplateMatcher, Region, to_image, HAS_RECOGNITION

IMAGES = Path(__file__).resolve().parent / ".." / "resources" / "images"

# TODO: Create new .png test data which works for pillow matching method
TEMPLATES = (
    ((26, 994, 274, 1115), "locator_Calculator_ctrl_One.jpg"),
    ((279, 994, 527, 1115), "locator_Calculator_ctrl_Two.jpg"),
    ((533, 994, 781, 1115), "locator_Calculator_ctrl_Three.jpg"),
    ((26, 868, 274, 989), "locator_Calculator_ctrl_Four.jpg"),
    ((279, 868, 527, 989), "locator_Calculator_ctrl_Five.jpg"),
    ((533, 868, 781, 989), "locator_Calculator_ctrl_Six.jpg"),
    ((26, 742, 274, 863), "locator_Calculator_ctrl_Seven.jpg"),
    ((279, 742, 527, 863), "locator_Calculator_ctrl_Eight.jpg"),
    ((533, 742, 781, 863), "locator_Calculator_ctrl_Nine.jpg"),
    ((26, 1120, 274, 1241), "locator_Calculator_ctrl_Positive_Negative.jpg"),
    ((279, 1120, 527, 1240), "locator_Calculator_ctrl_Zero.jpg"),
    ((533, 1120, 781, 1240), "locator_Calculator_ctrl_Decimal_Separator.jpg"),
    ((26, 742, 780, 1240), "locator_Calculator_ctrl_Number_pad.jpg"),
    ((786, 616, 1034, 737), "locator_Calculator_ctrl_Divide_by.jpg"),
    ((786, 742, 1034, 863), "locator_Calculator_ctrl_Multiply_by.jpg"),
    ((786, 868, 1034, 988), "locator_Calculator_ctrl_Minus.jpg"),
    ((786, 994, 1034, 1115), "locator_Calculator_ctrl_Plus.jpg"),
    ((786, 1120, 1034, 1240), "locator_Calculator_ctrl_Equals.jpg"),
    ((786, 616, 1034, 1240), "locator_Calculator_ctrl_Standard_operators.jpg"),
)


def get_test_name(param):
    return param[1].split(".")[0]


@pytest.fixture(params=TEMPLATES, ids=get_test_name)
def region_and_template(request):
    yield request.param[0], request.param[1]


@pytest.mark.skipif(not HAS_RECOGNITION, reason="Test requires recognition")
def test_find_template(region_and_template):
    region, template = region_and_template
    region = Region(*region)

    library = Images()
    matches = library.find_template_in_image(
        image=IMAGES / "source.png",
        template=IMAGES / template,
        tolerance=0.8,
    )

    assert len(matches) == 1
    match = matches[0]
    assert match.center == region.center



#!/usr/bin/env python3
"""
Create test data for FilesLibrary acceptance tests
"""
import argparse
import os
from pathlib import Path


def touch(path):
    Path(path).touch()


def mkdir(path):
    os.makedirs(path)


def write(path, content):
    if isinstance(content, str):
        content = content.encode("utf-8")
    with open(path, "wb") as fd:
        fd.write(content)


def prepare():
    mkdir("subfolder/first")
    mkdir("subfolder/second")
    mkdir("another/first")
    mkdir("another/second")
    mkdir("empty")

    touch("emptyfile")
    touch("notemptyfile")
    touch("sorted1.test")
    touch("sorted2.test")
    touch("sorted3.test")
    touch("subfolder/first/stuff.ext")
    touch("subfolder/second/stuff.ext")
    touch("another/first/noext")
    touch("another/second/one")
    touch("another/second/two")
    touch("another/second/three")
    touch("another/second/.dotfile")

    write("notemptyfile", "some content here")
    write("somebytes", b"\x00\x66\x6f\x6f\x00")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "root", default=os.getcwd(), help="Root directory for mock files"
    )
    args = parser.parse_args()

    mkdir(args.root)
    os.chdir(args.root)

    prepare()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['RPA',
 'RPA.Browser',
 'RPA.Cloud',
 'RPA.Desktop',
 'RPA.Desktop.keywords',
 'RPA.Email',
 'RPA.Excel',
 'RPA.Outlook',
 'RPA.Robocloud',
 'RPA.Robocorp',
 'RPA.Word',
 'RPA.includes',
 'RPA.scripts',
 'tests',
 'tests.python',
 'tests.resources',
 'tests.scripts']

package_data = \
{'': ['*'],
 'tests': ['robot/*'],
 'tests.resources': ['emails/*',
                     'excels/*',
                     'images/*',
                     'test-app/*',
                     'test-app/META-INF/*',
                     'work-items/__mail.html',
                     'work-items/__mail.html',
                     'work-items/__mail.html',
                     'work-items/__mail.html',
                     'work-items/__mail.html',
                     'work-items/items.json',
                     'work-items/items.json',
                     'work-items/items.json',
                     'work-items/items.json',
                     'work-items/items.json',
                     'work-items/mail-json.txt',
                     'work-items/mail-json.txt',
                     'work-items/mail-json.txt',
                     'work-items/mail-json.txt',
                     'work-items/mail-json.txt',
                     'work-items/mail-text.txt',
                     'work-items/mail-text.txt',
                     'work-items/mail-text.txt',
                     'work-items/mail-text.txt',
                     'work-items/mail-text.txt',
                     'work-items/mail-yaml.txt',
                     'work-items/mail-yaml.txt',
                     'work-items/mail-yaml.txt',
                     'work-items/mail-yaml.txt',
                     'work-items/mail-yaml.txt']}

install_requires = \
['O365>=2.0.21,<3.0.0',
 'PySocks>=1.5.6,!=1.5.7,<2.0.0',
 'PyYAML>=5.4.1,<6.0.0',
 'chardet>=3.0.0,<4.0.0',
 'click>=8.1.2,<9.0.0',
 'cryptography>=37.0.4,<38.0.0',
 'docutils',
 'exchangelib>=4.7.6,<5.0.0',
 'graphviz>=0.13.2,<0.14.0',
 'htmldocx>=0.0.6,<0.0.7',
 'hubspot-api-client>=4.0.6,<5.0.0',
 'importlib-metadata>=4.13.0,<5.0.0',
 'java-access-bridge-wrapper>=0.9.5,<0.10.0',
 'jsonpath-ng>=1.5.2,<2.0.0',
 'mss>=6.0.0,<7.0.0',
 'netsuitesdk>=1.1.0,<2.0.0',
 'notifiers>=1.3.3,<2.0.0',
 'openpyxl>=3.0.9,<4.0.0',
 'pillow>=9.1.1,<10.0.0',
 'pynput-robocorp-fork>=4.0.0,<5.0.0',
 'pyotp>=2.6.0,<3.0.0',
 'pyperclip>=1.8.0,<2.0.0',
 'robotframework-pythonlibcore>=3.0.0,<4.0.0',
 'robotframework-requests>=0.9.1,<0.10.0',
 'robotframework-seleniumlibrary>=6.0.0,<7.0.0',
 'robotframework-seleniumtestability>=2.0.0,<3.0.0',
 'robotframework>=4.0.0,!=4.0.1,<6.0.0',
 'rpaframework-core>=10.0.1,<11.0.0',
 'rpaframework-dialogs>=4.0.0,<5.0.0',
 'rpaframework-pdf>=5.0.0,<6.0.0',
 'selenium>=4.4.0,<5.0.0',
 'simple_salesforce>=1.0.0,<2.0.0',
 'tenacity>=8.0.1,<9.0.0',
 'tweepy>=3.8.0,<4.0.0',
 'tzlocal>=4.2,<5.0',
 'xlrd>=2.0.1,<3.0.0',
 'xlutils>=2.0.0,<3.0.0',
 'xlwt>=1.3.0,<2.0.0']

extras_require = \
{':python_full_version != "3.7.6" and python_full_version != "3.8.1" and sys_platform == "win32"': ['pywinauto>=0.6.8,<0.7.0',
                                                                                                    'pywin32>=302,<304'],
 ':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.7,<0.8'],
 ':sys_platform == "linux"': ['python-xlib>=0.17'],
 ':sys_platform == "win32"': ['rpaframework-windows>=6.1.0,<7.0.0',
                              'robotframework-sapguilibrary>=1.1,<2.0',
                              'comtypes==1.1.11',
                              'psutil>=5.7.0,<6.0.0'],
 'aws': ['rpaframework-aws>=5.1.0,<6.0.0'],
 'cv': ['rpaframework-recognition>=5.0.0,<6.0.0'],
 'playwright:python_version >= "3.7" and python_version < "4.0"': ['robotframework-browser>=13.0.0,<14.0.0']}

entry_points = \
{'console_scripts': ['rpa-crypto = RPA.scripts.crypto:main',
                     'use-robocorp-vault = RPA.scripts.robocorp_cloud:main']}

setup_kwargs = {
    'name': 'rpaframework',
    'version': '17.6.0',
    'description': 'A collection of tools and libraries for RPA',
    'long_description': 'RPA Framework\n=============\n\n.. contents:: Table of Contents\n   :local:\n   :depth: 1\n\n.. include-marker\n\nIntroduction\n------------\n\n`RPA Framework` is a collection of open-source libraries and tools for\nRobotic Process Automation (RPA), and it is designed to be used with both\n`Robot Framework`_ and `Python`_. The goal is to offer well-documented and\nactively maintained core libraries for Software Robot Developers.\n\nLearn more about RPA at `Robocorp Documentation`_.\n\n**The project is:**\n\n- 100% Open Source\n- Sponsored by Robocorp_\n- Optimized for Robocorp `Control Room`_ and `Developer Tools`_\n- Accepting external contributions\n\n.. _Robot Framework: https://robotframework.org\n.. _Robot Framework Foundation: https://robotframework.org/foundation/\n.. _Python: https://www.python.org/\n.. _Robocorp: https://robocorp.com\n.. _Robocorp Documentation: https://robocorp.com/docs/\n.. _Control Room: https://robocorp.com/docs/control-room\n.. _Developer Tools: https://robocorp.com/downloads\n.. _Installing Python Packages: https://robocorp.com/docs/setup/installing-python-package-dependencies\n\nLinks\n^^^^^\n\n- Homepage: `<https://www.github.com/robocorp/rpaframework/>`_\n- Documentation: `<https://rpaframework.org/>`_\n- PyPI: `<https://pypi.org/project/rpaframework/>`_\n- Release notes: `<https://rpaframework.org/releasenotes.html>`_\n- RSS feed: `<https://rpaframework.org/releases.xml>`_\n\n------------\n\n.. image:: https://img.shields.io/github/workflow/status/robocorp/rpaframework/main?style=for-the-badge\n   :target: https://github.com/robocorp/rpaframework/actions?query=workflow%3Amain\n   :alt: Status\n\n.. image:: https://img.shields.io/pypi/dw/rpaframework?style=for-the-badge\n   :target: https://pypi.python.org/pypi/rpaframework\n   :alt: rpaframework\n\n.. image:: https://img.shields.io/pypi/l/rpaframework.svg?style=for-the-badge&color=brightgreen\n   :target: http://www.apache.org/licenses/LICENSE-2.0.html\n   :alt: License\n\n------------\n\nPackages\n--------\n\n.. image:: https://img.shields.io/pypi/v/rpaframework.svg?label=rpaframework&style=for-the-badge\n   :target: https://pypi.python.org/pypi/rpaframework\n   :alt: rpaframework latest version\n\n\n\n.. image:: https://img.shields.io/pypi/v/rpaframework-aws.svg?label=rpaframework-aws&style=for-the-badge\n   :target: https://pypi.python.org/pypi/rpaframework-aws\n   :alt: rpaframework-aws latest version\n\n\n\n.. image:: https://img.shields.io/pypi/v/rpaframework-core.svg?label=rpaframework-core&style=for-the-badge\n   :target: https://pypi.python.org/pypi/rpaframework-core\n   :alt: rpaframework-core latest version\n\n\n\n.. image:: https://img.shields.io/pypi/v/rpaframework-dialogs.svg?label=rpaframework-dialogs&style=for-the-badge&color=blue\n   :target: https://pypi.python.org/pypi/rpaframework-dialogs\n   :alt: rpaframework-dialogs latest version\n\n\n\n.. image:: https://img.shields.io/pypi/v/rpaframework-google.svg?label=rpaframework-google&style=for-the-badge&color=blue\n   :target: https://pypi.python.org/pypi/rpaframework-google\n   :alt: rpaframework-google latest version\n\n\n\n.. image:: https://img.shields.io/pypi/v/rpaframework-pdf.svg?label=rpaframework-pdf&style=for-the-badge&color=blue\n   :target: https://pypi.python.org/pypi/rpaframework-pdf\n   :alt: rpaframework-pdf latest version\n\n\n\n.. image:: https://img.shields.io/pypi/v/rpaframework-recognition.svg?label=rpaframework-recognition&style=for-the-badge&color=blue\n   :target: https://pypi.python.org/pypi/rpaframework-recognition\n   :alt: rpaframework-recognition latest version\n\n\n\n.. image:: https://img.shields.io/pypi/v/rpaframework-windows.svg?label=rpaframework-windows&style=for-the-badge&color=blue\n   :target: https://pypi.python.org/pypi/rpaframework-windows\n   :alt: rpaframework-windows latest version\n\nFrom the above packages **rpaframework-core** and **rpaframework-recognition** are support packages, which themselves do **not** contain any libraries.\n\n\nLibraries\n---------\n\nThe RPA Framework project currently includes the following libraries:\n\nThe ``x`` in the **PACKAGE** column means that library **is** included in the **rpaframework** package and for example. ``x,dialogs`` means that ``RPA.Dialogs`` library is provided in both the **rpaframework** and **rpaframework-dialogs** packages.\n\n+----------------------------+----------------------------------------------+------------------------+\n| **LIBRARY NAME**           | **DESCRIPTION**                              | **PACKAGE**            |\n+----------------------------+----------------------------------------------+------------------------+\n| `Archive`_                 | Archiving TAR and ZIP files                  | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Base64AI`_                | Intelligent Document Processing service      | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Browser.Selenium`_        | Control browsers and automate the web        | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Browser.Playwright`_      | Newer way to control browsers                | special (more below)   |\n+----------------------------+----------------------------------------------+------------------------+\n| `Cloud.AWS`_               | Use Amazon AWS services                      | x,aws                  |\n+----------------------------+----------------------------------------------+------------------------+\n| `Cloud.Azure`_             | Use Microsoft Azure services                 | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Cloud.Google`_            | Use Google Cloud services                    | google                 |\n+----------------------------+----------------------------------------------+------------------------+\n| `Crypto`_                  | Common hashing and encryption operations     | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Database`_                | Interact with databases                      | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Desktop`_                 | Cross-platform desktop automation            | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Desktop.Clipboard`_       | Interact with the system clipboard           | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Desktop.OperatingSystem`_ | Read OS information and manipulate processes | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Desktop.Windows`_         | Automate Windows desktop applications        | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Dialogs`_                 | Request user input during executions         | x,dialogs              |\n+----------------------------+----------------------------------------------+------------------------+\n| `Email.Exchange`_          | E-Mail operations (Exchange protocol)        | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Email.ImapSmtp`_          | E-Mail operations (IMAP & SMTP)              | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Excel.Application`_       | Control the Excel desktop application        | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Excel.Files`_             | Manipulate Excel files directly              | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `FileSystem`_              | Read and manipulate files and paths          | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `FTP`_                     | Interact with FTP servers                    | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `HTTP`_                    | Interact directly with web APIs              | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Hubspot`_                 | Access HubSpot CRM data objects              | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Images`_                  | Manipulate images                            | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `JavaAccessBridge`_        | Control Java applications                    | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `JSON`_                    | Manipulate JSON objects                      | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `MFA`_                     | Authenticate using one-time passwords (OTP)  | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Nanonets`_                | Intelligent Document Processing service      | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Notifier`_                | Notify messages using different services     | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Outlook.Application`_     | Control the Outlook desktop application      | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `PDF`_                     | Read and create PDF documents                | x,pdf                  |\n+----------------------------+----------------------------------------------+------------------------+\n| `Robocorp.Process`_        | Use the Robocorp Process API                 | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Robocorp.WorkItems`_      | Use the Robocorp Work Items API              | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Robocorp.Vault`_          | Use the Robocorp Secrets API                 | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Salesforce`_              | Salesforce operations                        | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `SAP`_                     | Control SAP GUI desktop client               | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Tables`_                  | Manipulate, sort, and filter tabular data    | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Tasks`_                   | Control task execution                       | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Twitter`_                 | Twitter API interface                        | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n| `Windows`_                 | Alternative library for Windows automation   | x,windows              |\n+----------------------------+----------------------------------------------+------------------------+\n| `Word.Application`_        | Control the Word desktop application         | x                      |\n+----------------------------+----------------------------------------------+------------------------+\n\n.. _Archive: https://rpaframework.org/libraries/archive/\n.. _Base64AI: https://rpaframework.org/libraries/base64ai/\n.. _Browser.Playwright: https://rpaframework.org/libraries/browser_playwright/\n.. _Browser.Selenium: https://rpaframework.org/libraries/browser_selenium/\n.. _Cloud.AWS: https://rpaframework.org/libraries/cloud_aws/\n.. _Cloud.Azure: https://rpaframework.org/libraries/cloud_azure/\n.. _Cloud.Google: https://rpaframework.org/libraries/cloud_google/\n.. _Crypto: https://rpaframework.org/libraries/crypto/\n.. _Database: https://rpaframework.org/libraries/database/\n.. _Desktop: https://rpaframework.org/libraries/desktop/\n.. _Desktop.Clipboard: https://rpaframework.org/libraries/desktop_clipboard/\n.. _Desktop.Operatingsystem: https://rpaframework.org/libraries/desktop_operatingsystem/\n.. _Desktop.Windows: https://rpaframework.org/libraries/desktop_windows/\n.. _Dialogs: https://rpaframework.org/libraries/dialogs/\n.. _Email.Exchange: https://rpaframework.org/libraries/email_exchange/\n.. _Email.ImapSmtp: https://rpaframework.org/libraries/email_imapsmtp/\n.. _Excel.Application: https://rpaframework.org/libraries/excel_application/\n.. _Excel.Files: https://rpaframework.org/libraries/excel_files/\n.. _FileSystem: https://rpaframework.org/libraries/filesystem/\n.. _FTP: https://rpaframework.org/libraries/ftp/\n.. _HTTP: https://rpaframework.org/libraries/http/\n.. _Hubspot: https://rpaframework.org/libraries/hubspot/\n.. _Images: https://rpaframework.org/libraries/images/\n.. _JavaAccessBridge: https://rpaframework.org/libraries/javaaccessbridge/\n.. _JSON: https://rpaframework.org/libraries/json/\n.. _MFA: https://rpaframework.org/libraries/mfa/\n.. _Nanonets: https://rpaframework.org/libraries/nanonets/\n.. _Notifier: https://rpaframework.org/libraries/notifier/\n.. _Outlook.Application: https://rpaframework.org/libraries/outlook_application/\n.. _PDF: https://rpaframework.org/libraries/pdf/\n.. _Robocorp.Process: https://rpaframework.org/libraries/robocorp_process/\n.. _Robocorp.WorkItems: https://rpaframework.org/libraries/robocorp_workitems/\n.. _Robocorp.Vault: https://rpaframework.org/libraries/robocorp_vault/\n.. _Salesforce: https://rpaframework.org/libraries/salesforce/\n.. _SAP: https://rpaframework.org/libraries/sap/\n.. _Tables: https://rpaframework.org/libraries/tables/\n.. _Tasks: https://rpaframework.org/libraries/tasks/\n.. _Twitter: https://rpaframework.org/libraries/twitter/\n.. _Windows: https://rpaframework.org/libraries/windows/\n.. _Word.Application: https://rpaframework.org/libraries/word_application/\n\nInstallation of RPA.Browser.Playwright\n--------------------------------------\n\nThe **RPA.Browser.Playwright** at the moment requires special installation, because\nof the package size and the post install step it needs to be fully installed.\n\nMinimum required conda.yaml to install Playwright:\n\n.. code-block:: yaml\n\n   channels:\n     - conda-forge\n   dependencies:\n     - python=3.9.13\n     - pip=22.1.2\n     - nodejs=16.4.2\n     - pip:\n       - robotframework-browser==13.3.0\n       - rpaframework==17.0.1\n   rccPostInstall:\n     - rfbrowser init\n\nInstallation\n------------\n\nLearn about installing Python packages at `Installing Python Packages`_.\n\nDefault installation method with Robocorp `Developer Tools`_ using conda.yaml:\n\n.. code-block:: yaml\n\n   channels:\n     - conda-forge\n   dependencies:\n     - python=3.9.13\n     - pip=22.1.2\n     - pip:\n       - rpaframework==17.0.1\n\nTo install all extra packages (including Playwright dependencies), you can use:\n\n.. code-block:: yaml\n\n   channels:\n     - conda-forge\n   dependencies:\n     - python=3.9.13\n     - tesseract=4.1.1\n     - pip=22.1.2\n     - nodejs=16.14.2\n     - pip:\n       - robotframework-browser==13.3.0\n       - rpaframework==17.0.1   # rpaframework[aws]==13.0.0\n       - rpaframework-aws==5.0.0\n       - rpaframework-google==6.1.1\n       - rpaframework-recognition==5.0.0\n   rccPostInstall:\n     - rfbrowser init\n\nSeparate installation of AWS, Dialogs, PDF and Windows libraries without main rpaframework:\n\n.. code-block:: yaml\n\n   channels:\n     - conda-forge\n   dependencies:\n     - python=3.9.13\n     - pip=22.1.2\n     - pip:\n       - rpaframework-aws==5.0.0 # included in the rpaframework as an extra\n       - rpaframework-dialogs==4.0.0  # included in the rpaframework by default\n       - rpaframework-pdf==5.0.0  # included in the rpaframework by default\n       - rpaframework-windows==6.0.1 # included in the rpaframework by default\n\n\n.. note:: Python 3.7 or higher is required\n\nExample\n-------\n\nAfter installation the libraries can be directly imported inside\n`Robot Framework`_:\n\n.. code:: robotframework\n\n    *** Settings ***\n    Library    RPA.Browser.Selenium\n\n    *** Tasks ***\n    Login as user\n        Open available browser    https://example.com\n        Input text    id:user-name    ${USERNAME}\n        Input text    id:password     ${PASSWORD}\n\nThe libraries are also available inside Python_:\n\n.. code:: python\n\n    from RPA.Browser.Selenium import Selenium\n\n    lib = Selenium()\n\n    lib.open_available_browser("https://example.com")\n    lib.input_text("id:user-name", username)\n    lib.input_text("id:password", password)\n\nSupport and contact\n-------------------\n\n- `rpaframework.org <https://rpaframework.org/>`_ for library documentation\n- `Robocorp Documentation`_ for guides and tutorials\n- **#rpaframework** channel in `Robot Framework Slack`_ if you\n  have open questions or want to contribute\n- `Robocorp Forum`_ for discussions about RPA\n- Communicate with your fellow Software Robot Developers and Robocorp experts\n  at `Robocorp Developers Slack`_\n\n.. _Robot Framework Slack: https://robotframework-slack-invite.herokuapp.com/\n.. _Robocorp Forum: https://forum.robocorp.com\n.. _Robocorp Developers Slack: https://robocorp-developers.slack.com\n\nContributing\n------------\n\nFound a bug? Missing a critical feature? Interested in contributing?\nHead over to the `Contribution guide <https://rpaframework.org/contributing/guide.html>`_\nto see where to get started.\n\nDevelopment\n-----------\n\nRepository development is `Python`_ based and requires at minimum\nPython version 3.7+ installed on the development machine. The default Python version used in the\nRobocorp Robot template is 3.7.5 so it is a good choice for the version to install. Not recommended\nversions are 3.7.6 and 3.8.1, because they have issues with some of the dependencies related to rpaframework.\nAt the time the newer Python versions starting from 3.9 are also not recommended, because some of\nthe dependencies might cause issues.\n\nRepository development tooling is based on basically on `poetry`_ and `invoke`_. Poetry is the\nunderlying tool used for compiling, building and running the package. Invoke is used for scripting\npurposes for example for linting, testing and publishing tasks.\n\nFirst steps to start developing:\n\n1. initial poetry configuration\n\n.. code:: shell\n\n   poetry config virtualenvs.path null\n   poetry config virtualenvs.in-project true\n   poetry config repositories.devpi "https://devpi.robocorp.cloud/ci/test"\n\n2. git clone the repository\n#. create a new Git branch or switch to correct branch or stay in master branch\n\n   - some branch naming conventions **feature/name-of-feature**, **hotfix/name-of-the-issue**, **release/number-of-release**\n\n#. ``poetry install`` which install package with its dependencies into the **.venv** directory of the package, for example **packages/main/.venv**\n#. if testing against Robocorp Robot which is using **devdata/env.json**\n\n   - set environment variables\n   - or ``poetry build`` and use resulting .whl file (in the **dist/** directory) in the Robot **conda.yaml**\n   - or ``poetry build`` and push resulting .whl file  (in the **dist/** directory) into a repository and use raw url\n     to include it in the Robot **conda.yaml**\n   - another possibility for Robocorp internal development is to use Robocorp **devpi** instance, by ``poetry publish --ci``\n     and point **conda.yaml** to use rpaframework version in devpi\n\n#. ``poetry run python -m robot <ROBOT_ARGS> <TARGET_ROBOT_FILE>``\n\n   - common *ROBOT_ARGS* from Robocorp Robot template: ``--report NONE --outputdir output --logtitle "Task log"``\n\n#. ``poetry run python <TARGET_PYTHON_FILE>``\n#. ``invoke lint`` to make sure that code formatting is according to **rpaframework** repository guidelines.\n   It is possible and likely that Github action will fail the if developer has not linted the code changes. Code \n   formatting is based on `black`_ and `flake8`_ and those are run with the ``invoke lint``.\n#. the library documentation can be created in the repository root (so called "meta" package level). The documentation is \n   built by the docgen tools using the locally installed version of the project, local changes for the main package\n   will be reflected each time you generate the docs, but if you want to see local changes for optional packages, you must\n   utilize ``invoke install-local --package <package_name>`` using the appropriate package name (e.g., ``rpaframework-aws``). This\n   will reinstall that package as a local editable version instead of from PyPI. Multiple such packages can be added by\n   repeating the use of the ``--package`` option. In order to reset this, use ``invoke install --reset``.\n\n   - ``poetry update`` and/or ``invoke install-local --package <package name>``\n   - ``make docs``\n   - open ``docs/build/html/index.html`` with the browser to view the changes or execute ``make local`` and navigate to\n     ``localhost:8000`` to view docs as a live local webpage.\n\n   .. code-block:: toml\n\n      # Before\n      [tool.poetry.dependencies]\n      python = "^3.7"\n      rpaframework = { path = "packages/main", extras = ["cv", "playwright", "aws"] }\n      rpaframework-google = "^4.0.0"\n      rpaframework-windows = "^4.0.0"\n\n      # After\n      [tool.poetry.dependencies]\n      python = "^3.7"\n      rpaframework = { path = "packages/main", extras = ["cv", "playwright"] }\n      rpaframework-aws = { path = "packages/aws" }\n      rpaframework-google = "^4.0.0"\n      rpaframework-windows = "^4.0.0"\n\n#. ``invoke test`` (this will run both Python unittests and robotframework tests defined in the packages **tests/ directory**)\n\n   - to run specific Python test: ``poetry run pytest path/to/test.py::test_function``\n   - to run specific Robotframework test: ``inv testrobot -r <robot_name> -t <task_name>``\n\n#. git commit changes\n#. git push changes to remote\n#. create pull request from the branch describing changes included in the description\n#. update **docs/source/releasenotes.rst** with changes (commit and push)\n\nPackaging and publishing are done after changes have been merged into master branch.\nAll the following steps should be done within master branch.\n\n#. git pull latest changes into master branch\n#. in the package directory containing changes execute ``invoke lint`` and ``invoke test``\n#. update **pyproject.toml** with new version according to semantic versioning\n#. update **docs/source/releasenotes.rst** with changes\n#. in the repository root (so called "meta" package level) run command ``poetry update``\n#. git commit changed **poetry.lock** files (on meta and target package level), **releasenotes.rst**\n   and **pyproject.toml** with message "PACKAGE. version x.y.z"\n#. git push\n#. ``invoke publish`` after Github action on master branch is all green\n\nSome recommended tools for development\n\n- `Visual Studio Code`_ as a code editor with following extensions:\n\n   - `Robocorp Code`_\n   - `Robot Framework Language Server`_\n   - `GitLens`_\n   - `Python extension`_\n\n- `GitHub Desktop`_ will make version management less prone to errors\n\n.. _poetry: https://python-poetry.org\n.. _invoke: https://www.pyinvoke.org\n.. _Visual Studio Code: https://code.visualstudio.com\n.. _GitHub Desktop: https://desktop.github.com\n.. _Robocorp Code: https://marketplace.visualstudio.com/items?itemName=robocorp.robocorp-code\n.. _Robot Framework Language Server: https://marketplace.visualstudio.com/items?itemName=robocorp.robotframework-lsp\n.. _GitLens: https://marketplace.visualstudio.com/items?itemName=eamodio.gitlens\n.. _Python extension: https://marketplace.visualstudio.com/items?itemName=ms-python.python\n.. _black: https://pypi.org/project/black/\n.. _flake8: https://pypi.org/project/flake8/\n\nLicense\n-------\n\nThis project is open-source and licensed under the terms of the\n`Apache License 2.0 <http://apache.org/licenses/LICENSE-2.0>`_.\n',
    'author': 'RPA Framework',
    'author_email': 'rpafw@robocorp.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://rpaframework.org/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import logging
import platform
import warnings
from typing import Optional

if platform.system() == "Windows":
    # Configure comtypes to not generate DLL bindings into
    # current environment, instead keeping them in memory.
    # Slower, but prevents dirtying environments.
    import comtypes.client

    comtypes.client.gen_dir = None

    # Ignore pywinauto warning about threading mode,
    # which comtypes initializes to STA instead of MTA on import.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        import pywinauto

# pylint: disable=wrong-import-position
from robotlibcore import DynamicCore
from RPA.Desktop.utils import Buffer
from RPA.Desktop.keywords import (
    ElementNotFound,
    MultipleElementsFound,
    TimeoutException,
    ApplicationKeywords,
    ClipboardKeywords,
    FinderKeywords,
    KeyboardKeywords,
    MouseKeywords,
    ScreenKeywords,
    TextKeywords,
)


class Desktop(DynamicCore):
    """`Desktop` is a cross-platform library for navigating and interacting with
    desktop environments. It can be used to automate applications through
    the same interfaces that are available to human users.

    The library includes the following features:

    - Mouse and keyboard input emulation
    - Starting and stopping applications
    - Finding elements through image template matching
    - Scraping text from given regions
    - Taking screenshots
    - Clipboard management

    .. warning:: Windows element selectors are not currently supported, and require the use of ``RPA.Desktop.Windows``

    **Installation**

    The basic features such as mouse and keyboard input and application
    control work with a default ``rpaframework`` install.

    Advanced computer-vision features such as image template matching and
    OCR require an additional library called ``rpaframework-recognition``.

    The dependency can be either added separately or through additional
    extras with ``rpaframework[cv]``. If installing recognition through
    ``pip`` instead of ``conda``, the OCR feature also requires ``tesseract``.

    **Locating elements**

    To automate actions on the desktop, a robot needs to interact with various
    graphical elements such as buttons or input fields. The locations of these
    elements can be found using a feature called `locators`.

    A locator describes the properties or features of an element. This information
    can be later used to locate similar elements even when window positions or
    states change.

    The currently supported locator types are:

    =========== ================================================ ===========
    Name        Arguments                                        Description
    =========== ================================================ ===========
    alias       name (str)                                       A custom named locator from the locator database, the default.
    image       path (str)                                       Image of an element that is matched to current screen content.
    point       x (int), y (int)                                 Pixel coordinates as absolute position.
    offset      x (int), y (int)                                 Pixel coordinates relative to current mouse position.
    size        width (int), height (int)                        Region of fixed size, around point or screen top-left
    region      left (int), top (int), right (int), bottom (int) Bounding coordinates for a rectangular region.
    ocr         text (str), confidence (float, optional)         Text to find from the current screen.
    =========== ================================================ ===========

    A locator is defined by its type and arguments, divided by a colon.
    Some example usages are shown below. Note that the prefix for ``alias`` can
    be omitted as its the default type.

    .. code-block:: robotframework

        Click       point:50,100
        Click       region:20,20,100,30

        Move mouse  image:%{ROBOT_ROOT}/logo.png
        Move mouse  offset:200,0
        Click

        Click       alias:SpareBin.Login
        Click       SpareBin.Login

        Click       ocr:"Create New Account"

    You can also pass internal ``region`` objects as locators:

    .. code-block:: robotframework

        ${region}=  Find Element  ocr:"Customer name"
        Click       ${region}

    **Locator chaining**

    Often it is not enough to have one locator, but instead an element
    is defined through a relationship of various locators. For this use
    case the library supports a special syntax, which we will call
    locator chaining.

    An example of chaining:

    .. code-block:: robotframework

        # Read text from area on the right side of logo
        Read text    image:logo.png + offset:600,0 + size:400,200

    The supported operators are:

    ========== =========================================
    Operator   Description
    ========== =========================================
    then, +    Base locator relative to the previous one
    and, &&, & Both locators should be found
    or, ||, |  Either of the locators should be found
    not, !     The locator should not be found
    ========== =========================================

    Further examples:

    .. code-block:: robotframework

        # Click below either label
        Click    (image:name.png or image:email.png) then offset:0,300

        # Wait until dialog disappears
        Wait for element    not image:cookie.png

    **Named locators**

    The library supports storing locators in a database, which contains
    all of the required fields and various bits of metadata. This enables
    having one source of truth, which can be updated if a website's or applications's
    UI changes. Robot Framework scripts can then only contain a reference
    to a stored locator by name.

    The main way to create named locators is with `VSCode`_.

    Read more on identifying elements and crafting locators:

    - `Desktop automation and RPA <https://robocorp.com/docs/development-guide/desktop>`_
    - `How to find user interface elements using locators and keyboard shortcuts in Windows applications <https://robocorp.com/docs/development-guide/desktop/how-to-find-user-interface-elements-using-locators-and-keyboard-shortcuts-in-windows-applications>`_

    .. _VSCode: https://robocorp.com/docs/developer-tools/visual-studio-code/overview

    **Keyboard and mouse**

    Keyboard keywords can emulate typing text, but also pressing various function keys.
    The name of a key is case-insensitive and spaces will be converted to underscores,
    i.e. the key ``Page Down`` and ``page_down`` are equivalent.

    The following function keys are supported:

    =============== ===========
    Key             Description
    =============== ===========
    shift           A generic Shift key. This is a modifier.
    shift_l         The left Shift key. This is a modifier.
    shift_r         The right Shift key. This is a modifier.
    ctrl            A generic Ctrl key. This is a modifier.
    ctrl_l          he left Ctrl key. This is a modifier.
    ctrl_r          The right Ctrl key. This is a modifier.
    alt             A generic Alt key. This is a modifier.
    alt_l           The left Alt key. This is a modifier.
    alt_r           The right Alt key. This is a modifier.
    alt_gr          The AltGr key. This is a modifier.
    cmd             A generic command button (Windows / Command / Super key). This may be a modifier.
    cmd_l           The left command button (Windows / Command / Super key). This may be a modifier.
    cmd_r           The right command button (Windows / Command / Super key). This may be a modifier.
    up              An up arrow key.
    down            A down arrow key.
    left            A left arrow key.
    right           A right arrow key.
    enter           The Enter or Return key.
    space           The Space key.
    tab             The Tab key.
    backspace       The Backspace key.
    delete          The Delete key.
    esc             The Esc key.
    home            The Home key.
    end             The End key.
    page_down       The Page Down key.
    page_up         The Page Up key.
    caps_lock       The Caps Lock key.
    f1 to f20       The function keys.
    insert          The Insert key. This may be undefined for some platforms.
    menu            The Menu key. This may be undefined for some platforms.
    num_lock        The Num Lock key. This may be undefined for some platforms.
    pause           The Pause / Break key. This may be undefined for some platforms.
    print_screen    The Print Screen key. This may be undefined for some platforms.
    scroll_lock     The Scroll Lock key. This may be undefined for some platforms.
    =============== ===========

    When controlling the mouse, there are different types of actions that can be
    done. Same formatting rules as function keys apply. They are as follows:

    ============ ===========
    Action       Description
    ============ ===========
    click        Click with left mouse button
    left_click   Click with left mouse button
    double_click Double click with left mouse button
    triple_click Triple click with left mouse button
    right_click  Click with right mouse button
    ============ ===========

    The supported mouse button types are ``left``, ``right``, and ``middle``.

    **Examples**

    Both Robot Framework and Python examples follow.

    The library must be imported first.

    .. code-block:: robotframework

        *** Settings ***
        Library    RPA.Desktop

    .. code-block:: python

        from RPA.Desktop import Desktop
        desktop = Desktop()

    The library can open applications and interact with them through
    keyboard and mouse events.

    .. code-block:: robotframework

        *** Keywords ***
        Write entry in accounting
            [Arguments]    ${entry}
            Open application    erp_client.exe
            Click         image:%{ROBOT_ROOT}/images/create.png
            Type text     ${entry}
            Press keys    ctrl    s
            Press keys    enter

    .. code-block:: python

        def write_entry_in_accounting(entry):
            desktop.open_application("erp_client.exe")
            desktop.click(f"image:{ROBOT_ROOT}/images/create.png")
            desktop.type_text(entry)
            desktop.press_keys("ctrl", "s")
            desktop.press_keys("enter")


    Targeting can be currently done using coordinates (absolute or relative),
    but using template matching is preferred.

    .. code-block:: robotframework

        *** Keywords ***
        Write to field
            [Arguments]  ${text}
            Move mouse   image:input_label.png
            Move mouse   offset:200,0
            Click
            Type text    ${text}
            Press keys   enter

    .. code-block:: python

        def write_to_field(text):
            desktop.move_mouse("image:input_label.png")
            desktop.move_mouse("offset:200,0")
            desktop.click()
            desktop.type_text(text)
            desktop.press_keys("enter")


    Elements can be found by text too.

    .. code-block:: robotframework

        *** Keywords ***
        Click New
            Click       ocr:New

    .. code-block:: python

        def click_new():
            desktop.click('ocr:"New"')


    It is recommended to wait for the elements to be visible before
    trying any interaction. You can also pass ``region`` objects as locators.

    .. code-block:: robotframework

        *** Keywords ***
        Click New
            ${region}=  Wait For element  ocr:New
            Click       ${region}

    .. code-block:: python

        def click_new():
            region = desktop.wait_for_element("ocr:New")
            desktop.click(region)

    Another way to find elements by offsetting from an anchor:

    .. code-block:: robotframework

        *** Keywords ***
        Type Notes
            [Arguments]        ${text}
            Click With Offset  ocr:Notes  500  0
            Type Text          ${text}

    .. code-block:: python

        def type_notes(text):
            desktop.click_with_offset("ocr:Notes", 500, 0)
            desktop.type_text(text)

    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self, locators_path: Optional[str] = None):
        self.locators_path = locators_path

        self.logger = logging.getLogger(__name__)
        self.buffer = Buffer(self.logger)

        # Register keyword libraries to LibCore
        libraries = [
            ApplicationKeywords(self),
            ClipboardKeywords(self),
            FinderKeywords(self),
            KeyboardKeywords(self),
            MouseKeywords(self),
            ScreenKeywords(self),
            TextKeywords(self),
        ]
        super().__init__(libraries)



import base64
import logging
import mimetypes
from typing import Dict, Hashable, List, Optional, Union

import requests

from RPA.RobotLogListener import RobotLogListener

JSONValue = Optional[Union[str, int, float, bool]]
JSONType = Union[Dict[Hashable, JSONValue], List[JSONValue], JSONValue]


class Nanonets:
    """Library to support `Nanonets <https://nanonets.com/>`_ service for intelligent document processing (IDP).

    Added on **rpaframework** version: 17.0.1

    Service supports identifying fields in the documents, which can be given to the
    service in multiple different file formats and via URL.

    **Robot Framework example usage**

    .. code-block:: robotframework

        *** Settings ***
        Library   RPA.Nanonets
        Library   RPA.Robocorp.Vault

        *** Tasks ***
        Identify document
            ${secrets}=   Get Secret  nanonets-auth
            Set Authorization    ${secrets}[apikey]
            ${result}=    Predict File
            ...  ${CURDIR}${/}files${/}eckero.jpg
            ...  ${secrets}[receipts-model-id]
            ${fields}=    Get Fields From Prediction Result    ${result}
            FOR    ${field}    IN    @{fields}
                Log To Console    Label:${field}[label] Text:${field}[ocr_text]
            END
            ${tables}=    Get Tables From Prediction Result    ${result}
            FOR    ${table}    IN    @{tables}
                FOR    ${rows}    IN    ${table}[rows]
                    FOR    ${row}    IN    @{rows}
                        ${cells}=    Evaluate    [cell['text'] for cell in $row]
                        Log To Console    ROW:${{" | ".join($cells)}}
                    END
                END
            END


    **Python example usage**

    .. code-block:: python

        from RPA.Nanonets import Nanonets
        from RPA.Robocorp.Vault import Vault

        secrets = Vault().get_secret("nanonets-auth")
        nanolib = Nanonets()
        nanolib.set_authorization(secrets["apikey"])
        result = nanolib.predict_file(file_to_scan, secrets["receipts-model-id"])
        fields = nanolib.get_fields_from_prediction_result(result)
        for field in fields:
            print(f"Label: {field['label']} Text: {field['ocr_text']}")
        tables = nanolib.get_tables_from_prediction_result(result)
        for table in tables:
            rpatable = Tables().create_table(table["rows"])
            for row in table["rows"]:
                cells = [cell["text"] for cell in row]
                print(f"ROW: {' | '.join(cells)}")
    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.base_url = "https://app.nanonets.com/api/v2"
        self._request_headers = {"Content-Type": "application/json"}
        self.apikey = None
        listener = RobotLogListener()
        listener.register_protected_keywords(["RPA.Nanonets.set_authorization"])

    def _get_file_base64_and_mimetype(self, file_path: str):
        with open(file_path, "rb") as image_file:
            encoded_content = base64.b64encode(image_file.read())
        return encoded_content.decode("utf-8"), mimetypes.guess_type(file_path)[0]

    def set_authorization(self, apikey: str) -> None:
        """Set Nanonets request headers with key related to API.

        :param apikey: key related to the API

        Robot Framework example:

        .. code-block:: robotframework

            ${secrets}=   Get Secret  nanonets-auth
            Set Authorization    ${secrets}[apikey]

        Python example:

        .. code-block:: python

            secrets = Vault().get_secret("nanonets-auth")
            nanolib = Nanonets()
            nanolib.set_authorization(secrets["apikey"])
        """
        self.apikey = apikey

    def ocr_fulltext(self, filename: str, filepath: str) -> List:
        """OCR fulltext a given file. Returns words and full text.

        Filename and filepath needs to be given separately.

        :param filename: name of the file
        :param filepath: path of the file
        :return: the result in a list format

        Robot Framework example:

        .. code-block:: robotframework

            ${results}=  OCR Fulltext
            ...   invoice.pdf
            ...   ${CURDIR}${/}invoice.pdf
            FOR  ${result}  IN  @{results}
                Log To Console  Filename: ${result}[filename]
                FOR  ${pagenum}  ${page}  IN ENUMERATE  @{result.pagedata}   start=1
                    Log To Console  Page ${pagenum} raw Text: ${page}[raw_text]
                END
            END

        Python example:

        .. code-block:: python

            results = nanolib.ocr_fulltext("IMG_8277.jpeg", "./IMG_8277.jpeg")
            for result in results:
                print(f"FILENAME: {result['filename']}")
                for page in result["page_data"]:
                    print(f"Page {page['page']+1}: {page['raw_text']}")
        """
        _, mime = self._get_file_base64_and_mimetype(filepath)

        # pylint: disable=R1732
        files = [("file", (filename, open(filepath, "rb"), mime))]

        response = requests.request(
            "POST",
            f"{self.base_url}/OCR/FullText",
            files=files,
            auth=requests.auth.HTTPBasicAuth(self.apikey, ""),
        )

        response.raise_for_status()
        return response.json()["results"]

    def get_all_models(self) -> Dict:
        """Get all available models related to the API key.

        :return: object containing available models

        Robot Framework example:

        .. code-block:: robotframework

            ${models}=  Get All Models
            FOR  ${model}  IN  @{models}
                Log To Console  Model ID: ${model}[model_id]
                Log To Console  Model Type: ${model}[model_type]
            END

        Python example:

        .. code-block:: python

            models = nanolib.get_all_models()
            for model in models:
                print(f"model id: {model['model_id']}")
                print(f"model type: {model['model_type']}")
        """
        response = requests.request(
            "GET",
            f"{self.base_url}/ImageCategorization/Models/",
            auth=requests.auth.HTTPBasicAuth(self.apikey, ""),
        )

        response.raise_for_status()
        return response.json()

    def predict_file(self, filepath: str, model_id: str) -> Dict:
        """Get prediction result for a file by a given model id.

        :param filepath: filepath to the file
        :param model_id: id of the Nanonets model to categorize a file
        :return: the result in a list format

        Robot Framework example:

        .. code-block:: robotframework

            ${result}=  Predict File  ./document.pdf   ${MODEL_ID}
            ${fields}=    Get Fields From Prediction Result    ${result}
            FOR    ${field}    IN    @{fields}
                Log To Console    Label:${field}[label] Text:${field}[ocr_text]
            END
            ${tables}=    Get Tables From Prediction Result    ${result}
            FOR    ${table}    IN    @{tables}
                FOR    ${rows}    IN    ${table}[rows]
                    FOR    ${row}    IN    @{rows}
                        ${cells}=    Evaluate    [cell['text'] for cell in $row]
                        Log To Console    ROW:${{" | ".join($cells)}}
                    END
                END
            END

        Python example:

        .. code-block:: python

            result = nanolib.predict_file("./docu.pdf", secrets["receipts-model-id"])
            fields = nanolib.get_fields_from_prediction_result(result)
            for field in fields:
                print(f"Label: {field['label']} Text: {field['ocr_text']}")
            tables = nanolib.get_tables_from_prediction_result(result)
            for table in tables:
                for row in table["rows"]:
                    cells = [cell["text"] for cell in row]
                    print(f"ROW: {' | '.join(cells)}")
        """

        url = f"{self.base_url}/OCR/Model/{model_id}/LabelFile/"

        # pylint: disable=R1732
        data = {"file": open(filepath, "rb")}

        response = requests.post(
            url,
            headers={},
            auth=requests.auth.HTTPBasicAuth(self.apikey, ""),
            files=data,
        )

        response.raise_for_status()
        return response.json()

    def get_fields_from_prediction_result(self, prediction: JSONType) -> List:
        """Helper keyword to get found fields from a prediction result.

        For example. see ``Predict File`` keyword

        :param prediction: prediction result file
        :return: list of found fields
        """
        return [
            item
            for item in prediction["result"][0]["prediction"]
            if "type" in item.keys() and item["type"] == "field"
        ]

    def get_tables_from_prediction_result(self, prediction: JSONType) -> List:
        """Helper keyword to get found tables from a prediction result.

        For another example. see ``Predict File`` keyword

        :param prediction: prediction result file
        :return: list of found tables

        Robot Framework example:

        .. code-block:: robotframework

            # It is possible to create ``RPA.Tables`` compatible tables from the result
            ${tables}=    Get Tables From Prediction Result    ${result}
            FOR    ${table}    IN    @{tables}
                ${rpatable}=    Create Table    ${table}[rows]
                FOR    ${row}    IN    @{rpatable}
                    Log To Console    ${row}
                END
            END

        Python example:

        .. code-block:: python

            # It is possible to create ``RPA.Tables`` compatible tables from the result
            tables = nanolib.get_tables_from_prediction_result(result)
            for table in tables:
                rpatable = Tables().create_table(table["rows"])
                for row in rpatable:
                    print(row)
        """
        tables = [
            item
            for item in prediction["result"][0]["prediction"]
            if "type" in item.keys() and item["type"] == "table"
        ]
        for table in tables:
            table["rows"] = []
            row = []
            last_row = 1
            for cell in table["cells"]:
                if len(row) == 0 or last_row == cell["row"]:
                    row.append(cell)
                elif last_row != cell["row"]:
                    table["rows"].append(row)
                    row = [cell]
                last_row = cell["row"]
            if len(row) > 0:
                table["rows"].append(row)
        return tables



"""Common utilities shared by any e-mail related library."""


from pathlib import Path


def counter_duplicate_path(file_path: Path) -> Path:
    """Returns a unique file path by adding a suffixed counter if already exists."""
    if not file_path.exists():
        return file_path  # unique already

    root_dir = file_path.parent
    duplicates = root_dir.glob(f"{file_path.stem}*{file_path.suffix}")
    suffixes = []
    for dup in duplicates:
        parts = dup.stem.rsplit("-", 1)
        if len(parts) == 2 and parts[1].isdigit():
            suffixes.append(int(parts[1]))
    next_suffix = max(suffixes) + 1 if suffixes else 2

    file_path = root_dir / f"{file_path.stem}-{next_suffix}{file_path.suffix}"
    return file_path



# pylint: disable=C0411,C0412,C0413
import logging
from pathlib import Path
from typing import Any, Optional, Union
from urllib.parse import urlparse

import RequestsLibrary.log
from RequestsLibrary.utils import is_file_descriptor
from robot.api import logger

from RPA.core.notebook import notebook_file
from RPA.FileSystem import FileSystem

LOG_CHAR_LIMIT = 10000


def log_response(response):
    logger.debug(
        "%s Response : url=%s \n " % (response.request.method.upper(), response.url)
        + "status=%s, reason=%s \n " % (response.status_code, response.reason)
        + "headers=%s \n " % response.headers
        + "body=%s \n " % format_data_to_log_string(response.text)
    )


def log_request(response):
    request = response.request
    if response.history:
        original_request = response.history[0].request
        redirected = "(redirected) "
    else:
        original_request = request
        redirected = ""
    logger.debug(
        "%s Request : " % original_request.method.upper()
        + "url=%s %s\n " % (original_request.url, redirected)
        + "path_url=%s \n " % original_request.path_url
        + "headers=%s \n " % original_request.headers
        + "body=%s \n " % format_data_to_log_string(original_request.body)
    )


def format_data_to_log_string(data, limit=LOG_CHAR_LIMIT):

    if not data:
        return None

    if is_file_descriptor(data):
        return repr(data)

    if len(data) > limit and logging.getLogger().level > 10:
        data = (
            "%s... (set the log level to DEBUG or TRACE to see the full content)"
            % data[:limit]
        )

    return data


RequestsLibrary.log.log_response = log_response
RequestsLibrary.log.log_request = log_request


from RequestsLibrary import RequestsLibrary  # noqa: E402

# NOTE. Above logging changes are related to. Especially on Automation Studio
# extensive INFO level logging makes readability problematic.
# https://github.com/MarketSquare/robotframework-requests/issues/353


class HTTP(RequestsLibrary):
    """The *RPA.HTTP* library extends functionality of the `RequestsLibrary`_.
    See that documentation for several examples of how to issue ``GET``
    requests and utilize the returned ``result`` objects.

    .. _RequestsLibrary: https://marketsquare.github.io/robotframework-requests/doc/RequestsLibrary.html

    This extension provides helper keywords to get an HTTP resource on a
    session. The ``HTTP Get`` and ``Download`` keywords will initiate a
    session if one does not exist for the provided URL, or use an existing
    session. When using these keywords, you do not need to manage
    sessions with ``Create Session``. Session management is still
    required if you use the underlying session keywords, e.g.,
    ``* On Session``.

    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "reST"

    def __init__(self, *args, **kwargs) -> None:
        RequestsLibrary.__init__(self, *args, **kwargs)
        self.logger = logging.getLogger(__name__)
        self.fs = FileSystem()
        self.session_alias_prefix = "rpasession_alias."
        self.current_session_alias = None

    def http_get(
        self,
        url: str,
        target_file: Optional[str] = None,
        verify: Union[bool, str] = True,
        force_new_session: bool = False,
        overwrite: bool = False,
        stream: bool = False,
        **kwargs,
    ) -> dict:
        """
        A helper method for ``Get Request`` that will create a session, perform GET
        request, and store the target file, if set by the ``target_file`` parameter.

        The old session will be used if the URL scheme and the host are the same as
        previously, e.g., 'https://www.google.fi' part of the URL.

        .. code-block:: robotframework

            *** Settings ***
            Library    RPA.HTTP

            *** Variables ***
            ${DOWNLOAD_PATH}=   ${OUTPUT DIR}${/}downloads
            ${WORD_EXAMPLE}=    https://file-examples.com/wp-content/uploads/2017/02/file-sample_100kB.doc
            ${EXCEL_EXAMPLE}=   https://file-examples.com/wp-content/uploads/2017/02/file_example_XLS_10.xls

            *** Tasks ***
            Download files with reused session
                # Starts a new session
                HTTP Get    ${WORD_EXAMPLE}    target_file=${DOWNLOAD_PATH}${/}word-example.doc
                # Uses the previous session
                HTTP Get    ${EXCEL_EXAMPLE}    target_file=${DOWNLOAD_PATH}${/}excel-example.xls

        :param url: target URL for GET request
        :param target_file: filepath to save request content, default ``None``
        :param verify: if SSL verification should be done, default ``True``,
            a CA_BUNDLE path can also be provided
        :param force_new_session: if new HTTP session should be created,
            default ``False``
        :param overwrite: used together with ``target_file``, if ``True`` will overwrite
            the target file, default ``False``
        :param stream: if ``False``, the response content will be immediately downloaded
        :return: request response as a dict
        """  # noqa: E501
        uc = urlparse(url)

        http_host = f"{uc.scheme}://{uc.netloc}"
        request_alias = f"{self.session_alias_prefix}{uc.scheme}{uc.netloc}"
        url_path = url.replace(http_host, "")

        if force_new_session or not self.session_exists(request_alias):
            self.logger.info("Creating a new HTTP session")
            self.create_session(request_alias, http_host, verify=verify)
        else:
            self.logger.info("Using already existing HTTP session")

        self.current_session_alias = request_alias
        response = self.get_on_session(request_alias, url_path, stream=stream, **kwargs)

        if target_file is not None:
            self._create_or_overwrite_target_file(
                target_file, response.content, overwrite
            )

        return response

    def _create_or_overwrite_target_file(
        self,
        path: str,
        response: Any,
        overwrite: bool,
    ) -> None:
        CHUNK_SIZE = 32768
        Path(path).parent.mkdir(parents=True, exist_ok=True)
        file_exists = Path(path).is_file()
        if not file_exists or (file_exists and overwrite):
            with open(path, "wb") as f:
                for chunk in response.iter_content(CHUNK_SIZE):
                    if chunk:  # filter out keep-alive new chunks
                        f.write(chunk)
        notebook_file(path)

    def get_current_session_alias(self) -> str:
        """Get request session alias that was used with the ``HTTP Get`` keyword.

        :return: name of session alias as a string
        """
        return self.current_session_alias

    def download(
        self,
        url: str,
        target_file: Optional[str] = None,
        verify: Union[bool, str] = True,
        force_new_session: bool = False,
        overwrite: bool = False,
        stream: bool = False,
        **kwargs,
    ) -> dict:
        """An alias for the ``HTTP Get`` keyword.

        The difference in use is that the URL is always downloaded based on
        the URL path (even without ``target_file``). If there is a filename
        in the path, then that is used as ``target_file`` to save to. By default,
        the filename will be "downloaded.html".

        .. code-block:: robotframework

            *** Settings ***
            Library    RPA.HTTP

            *** Variables ***
            ${DOWNLOAD_PATH}=   ${OUTPUT DIR}${/}downloads
            ${WORD_EXAMPLE}=    https://file-examples.com/wp-content/uploads/2017/02/file-sample_100kB.doc
            ${EXCEL_EXAMPLE}=   https://file-examples.com/wp-content/uploads/2017/02/file_example_XLS_10.xls

            *** Tasks ***
            Download files with reused session with provided file names
                # Starts a new session
                Download    ${WORD_EXAMPLE}    target_file=${DOWNLOAD_PATH}
                # Uses the previous session
                Download    ${EXCEL_EXAMPLE}    target_file=${DOWNLOAD_PATH}
                # Above files are downloaded using the same names as they have
                # on the remote server.

        :param url: target URL for GET request
        :param target_file: filepath to save request content, default ``None``
        :param verify: if SSL verification should be done, default ``True``,
         a CA_BUNDLE path can also be provided
        :param force_new_session: if new HTTP session should be created,
         default ``False``
        :param overwrite: used together with ``target_file``, if ``True`` will overwrite
         the target file, default ``False``
        :param stream: if ``False`` the response content will be immediately downloaded
        :return: request response as a dict
        """  # noqa: E501
        response = self.http_get(
            url,
            verify=verify,
            force_new_session=force_new_session,
            overwrite=overwrite,
            stream=stream,
            **kwargs,
        )

        dirname = Path()
        filename = None

        if target_file is not None:
            target = Path(target_file)
            if target.is_dir():
                dirname = target
            else:
                dirname = target.parent
                filename = target.name

        if filename is None:
            filename = urlparse(url).path.rsplit("/", 1)[-1] or "downloaded.html"

        self._create_or_overwrite_target_file(dirname / filename, response, overwrite)

        return response



import copy
import csv
import logging
import re
from collections import OrderedDict, namedtuple
from enum import Enum
from itertools import groupby, zip_longest
from keyword import iskeyword
from numbers import Number
from operator import itemgetter
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Union,
)

from robot.api.deco import keyword
from robot.libraries.BuiltIn import BuiltIn
from RPA.core.notebook import notebook_print, notebook_table
from RPA.core.types import is_dict_like, is_list_like, is_namedtuple


Index = Union[int, str]
Column = Union[int, str]
Row = Union[Dict, List, Tuple, NamedTuple, set]
Data = Optional[Union[Dict[Column, Row], List[Row], "Table"]]
CellCondition = Callable[[Any], bool]
RowCondition = Callable[[Union[Index, Row]], bool]


def to_list(obj: Any, size: int = 1):
    """Convert (possibly scalar) value to list of `size`."""
    if not is_list_like(obj):
        return [obj] * int(size)
    else:
        return obj


def to_identifier(val: Any):
    """Convert string to valid identifier"""
    val = str(val).strip()
    # Replaces spaces, dashes, and slashes to underscores
    val = re.sub(r"[\s\-/\\]", "_", val)
    # Remove remaining invalid characters
    val = re.sub(r"[^0-9a-zA-Z_]", "", val)
    # Identifier can't start with digits
    val = re.sub(r"^[^a-zA-Z_]+", "", val)

    if not val or iskeyword(val):
        raise ValueError(f"Unable to convert to identifier: {val}")

    return val


def to_condition(operator: str, value: Any) -> CellCondition:
    """Convert string operator into callable condition function."""
    operator = str(operator).lower().strip()
    condition = {
        ">": lambda x: x is not None and x > value,
        "<": lambda x: x is not None and x < value,
        ">=": lambda x: x is not None and x >= value,
        "<=": lambda x: x is not None and x <= value,
        "==": lambda x: x == value,
        "!=": lambda x: x != value,
        "is": lambda x: x is value,
        "not is": lambda x: x is not value,
        "contains": lambda x: x is not None and value in x,
        "not contains": lambda x: x is not None and value not in x,
        "in": lambda x: x in value,
        "not in": lambda x: x not in value,
    }.get(operator)

    if not condition:
        raise ValueError(f"Unknown operator: {operator}")

    return condition


def if_none(value: Any, default: Any):
    """Return default if value is None."""
    return value if value is not None else default


def uniq(seq: Iterable):
    """Return list of unique values while preserving order.
    Values must be hashable.
    """
    seen, result = {}, []
    for item in seq:
        if item in seen:
            continue
        seen[item] = None
        result.append(item)
    return result


class Dialect(Enum):
    """CSV dialect"""

    Excel = "excel"
    ExcelTab = "excel-tab"
    Unix = "unix"


class Table:
    """Container class for tabular data.

    Supported data formats:

    - empty: None values populated according to columns/index
    - list: list of data Rows
    - dict: Dictionary of columns as keys and Rows as values
    - table: An existing Table

    Row: a namedtuple, dictionary, list or a tuple

    :param data:     Values for table,  see "Supported data formats"
    :param columns:  Names for columns, should match data dimensions
    """

    def __init__(self, data: Data = None, columns: Optional[List[str]] = None):
        self._data = []
        self._columns = []

        # Use public setter to validate data
        if columns is not None:
            self.columns = list(columns)

        if not data:
            self._init_empty()
        elif isinstance(data, Table):
            self._init_table(data)
        elif is_dict_like(data):
            self._init_dict(data)
        elif is_list_like(data):
            self._init_list(data)
        else:
            raise TypeError("Not a valid input format")

        if columns:
            self._sort_columns(columns)

        self._validate_self()

    def _init_empty(self):
        """Initialize table with empty data."""
        self._data = []

    def _init_table(self, table: "Table"):
        """Initialize table with another table."""
        if not self.columns:
            self.columns = table.columns
        self._data = table.data

    def _init_list(self, data: List[Any]):
        """Initialize table from list-like container."""
        # Assume data is homogenous in regard to row type
        obj = data[0]
        column_names = self._column_name_getter(obj)
        column_values = self._column_value_getter(obj)

        # Map of source to destination column
        column_map = {}

        # Do not update columns if predefined
        add_columns = not bool(self._columns)

        for obj in data:
            row = [None] * len(self._columns)

            for column_src in column_names(obj):
                # Check if column has been added with different name
                column_dst = column_map.get(column_src, column_src)

                # Dictionaries and namedtuples can
                # contain unknown columns
                if column_dst not in self._columns:
                    if not add_columns:
                        continue

                    col = self._add_column(column_dst)

                    # Store map of source column name to created name
                    column_dst = self._columns[col]
                    column_map[column_src] = column_dst

                    while len(row) < len(self._columns):
                        row.append(None)

                col = self.column_location(column_dst)
                row[col] = column_values(obj, column_src)

            self._data.append(row)

    def _init_dict(self, data: Dict[Column, Row]):
        """Initialize table from dict-like container."""
        if not self._columns:
            self._columns = list(data.keys())

        # Filter values by defined columns
        columns = (
            to_list(values)
            for column, values in data.items()
            if column in self._columns
        )

        # Convert columns to rows
        self._data = [list(row) for row in zip_longest(*columns)]

    def __repr__(self):
        return "Table(columns={}, rows={})".format(self.columns, self.size)

    def __len__(self):
        return self.size

    def __iter__(self):
        return self.iter_dicts(with_index=False)

    def __eq__(self, other: Any):
        if not isinstance(other, Table):
            return False
        return self._columns == other._columns and self._data == other._data

    @property
    def data(self):
        return self._data.copy()

    @property
    def size(self) -> int:
        return len(self._data)

    @property
    def dimensions(self):
        return self.size, len(self._columns)

    @property
    def index(self):
        return list(range(self.size))

    @property
    def columns(self):
        return self._columns.copy()

    @columns.setter
    def columns(self, names):
        """Rename columns with given values."""
        self._validate_columns(names)
        self._columns = list(names)

    def _validate_columns(self, names):
        """Validate that given column names can be used."""
        if not is_list_like(names):
            raise ValueError("Columns should be list-like")

        if len(set(names)) != len(names):
            raise ValueError("Duplicate column names")

        if self._data and len(names) != len(self._data[0]):
            raise ValueError("Invalid columns length")

    def _column_name_getter(self, obj):
        """Create callable that returns column names for given obj types."""
        if is_namedtuple(obj):
            # Use namedtuple fields as columns
            def get(obj):
                return list(obj._fields)

        elif is_dict_like(obj):
            # Use dictionary keys as columns
            def get(obj):
                return list(obj.keys())

        elif is_list_like(obj):
            # Use either predefined columns, or
            # generate range-based column values
            predefined = list(self._columns)

            def get(obj):
                count = len(obj)
                if predefined:
                    if count > len(predefined):
                        raise ValueError(
                            f"Data had more than defined {len(predefined)} columns"
                        )
                    return predefined[:count]
                else:
                    return list(range(count))

        else:
            # Fallback to single column
            def get(_):
                return self._columns[:1] if self._columns else [0]

        return get

    def _column_value_getter(self, obj):
        """Create callable that returns column values for given object types."""
        if is_namedtuple(obj):
            # Get values using properties
            def get(obj, column):
                return getattr(obj, column, None)

        elif is_dict_like(obj):
            # Get values using dictionary keys
            def get(obj, column):
                return obj.get(column)

        elif is_list_like(obj):
            # Get values using list indexes
            def get(obj, column):
                col = self.column_location(column)
                try:
                    return obj[col]
                except IndexError:
                    return None

        else:
            # Fallback to single column
            def get(obj, _):
                return obj

        return get

    def _sort_columns(self, order):
        """Sort columns to match given order."""
        unknown = set(self._columns) - set(order)
        if unknown:
            names = ", ".join(str(name) for name in unknown)
            raise ValueError(f"Unknown columns: {names}")

        cols = [self.column_location(column) for column in order]

        self._columns = [self._columns[col] for col in cols]
        self._data = [[row[col] for col in cols] for row in self._data]

    def _validate_self(self):
        """Validate that internal data is valid and coherent."""
        self._validate_columns(self._columns)

        if self._data:
            head = self._data[0]
            if len(head) != len(self._columns):
                raise ValueError("Columns length does not match data")

    def index_location(self, value: Index) -> int:
        try:
            value = int(value)
        except ValueError as err:
            raise ValueError(f"Index is not a number: {value}") from err

        if value < 0:
            value += self.size

        if self.size == 0:
            raise IndexError("No rows in table")

        if (value < 0) or (value >= self.size):
            raise IndexError(f"Index ({value}) out of range (0..{self.size - 1})")

        return value

    def column_location(self, value):
        """Find location for column value."""

        # Try to use as-is
        try:
            return self._columns.index(value)
        except ValueError:
            pass

        # Try as integer index
        try:
            value = int(value)

            if value in self._columns:
                location = self._columns.index(value)
            elif value < 0:
                location = value + len(self._columns)
            else:
                location = value

            size = len(self._columns)
            if size == 0:
                raise IndexError("No columns in table")

            if location >= size:
                raise IndexError(f"Column ({location}) out of range (0..{size - 1})")

            return location
        except ValueError:
            pass

        # No matches
        options = ", ".join(str(col) for col in self._columns)
        raise ValueError(f"Unknown column name: {value}, current columns: {options}")

    def __getitem__(self, key):
        """Helper method for accessing items in the Table.

        Examples:
            table[:10]           First 10 rows
            table[0,1]           Value in first row and second column
            table[2:10,"email"]  Values in "email" column for rows 3 to 11
        """
        # Both row index and columns given
        if isinstance(key, tuple):
            index, column = key
            index = self._slice_index(index) if isinstance(index, slice) else index
            return self.get(indexes=index, columns=column, as_list=True)
        # Row indexed with slice, all columns
        elif isinstance(key, slice):
            return self.get(indexes=self._slice_index(key), as_list=True)
        # Single row
        else:
            return self.get(indexes=key, as_list=True)

    def __setitem__(self, key, value):
        """Helper method for setting items in the Table.

        Examples:
            table[5]  = ["Mikko", "Mallikas"]
            table[:2] = [["Marko", "Markonen"], ["Pentti", "Penttinen"]]
        """
        # Both row index and columns given
        if isinstance(key, tuple):
            index, column = key
            index = self._slice_index(index) if isinstance(index, slice) else index
            return self.set(indexes=index, columns=column, values=value)
        # Row indexed with slice, all columns
        elif isinstance(key, slice):
            return self.set(indexes=self._slice_index(key), values=value)
        # Single row
        else:
            return self.set(indexes=key, values=value)

    def _slice_index(self, slicer):
        """Create list of index values from slice object."""
        start = self.index_location(slicer.start) if slicer.start is not None else 0
        end = self.index_location(slicer.stop) if slicer.stop is not None else self.size
        return list(range(start, end))

    def copy(self):
        """Create a copy of this table."""
        return copy.deepcopy(self)

    def clear(self):
        """Remove all rows from this table."""
        self._data = []

    def head(self, rows, as_list=False):
        """Return first n rows of table."""
        indexes = self.index[: int(rows)]
        return self.get_table(indexes, as_list=as_list)

    def tail(self, rows, as_list=False):
        """Return last n rows of table."""
        indexes = self.index[-int(rows) :]
        return self.get_table(indexes, as_list=as_list)

    def get(self, indexes=None, columns=None, as_list=False):
        """Get values from table. Return type depends on input dimensions.

        If `indexes` and `columns` are scalar, i.e. not lists:
            Returns single cell value

        If either `indexes` or `columns` is a list:
            Returns matching row or column

        If both `indexes` and `columns` are lists:
            Returns a new Table instance with matching cell values

        :param indexes: List of indexes, or all if not given
        :param columns: List of columns, or all if not given
        """
        indexes = if_none(indexes, self.index)
        columns = if_none(columns, self._columns)

        if is_list_like(indexes) and is_list_like(columns):
            return self.get_table(indexes, columns, as_list)
        elif not is_list_like(indexes) and is_list_like(columns):
            return self.get_row(indexes, columns, as_list)
        elif is_list_like(indexes) and not is_list_like(columns):
            return self.get_column(columns, indexes, as_list)
        else:
            return self.get_cell(indexes, columns)

    def get_cell(self, index, column):
        """Get single cell value."""
        idx = self.index_location(index)
        col = self.column_location(column)

        return self._data[idx][col]

    def get_row(self, index: Index, columns=None, as_list=False):
        """Get column values from row.

        :param index:   Index for row
        :param columns: Column names to include, or all if not given
        :param as_list: Return row as dictionary, instead of list
        """
        columns = if_none(columns, self._columns)
        idx = self.index_location(index)

        if as_list:
            row = []
            for column in columns:
                col = self.column_location(column)
                row.append(self._data[idx][col])
            return row
        else:
            row = {}
            for column in columns:
                col = self.column_location(column)
                row[self._columns[col]] = self._data[idx][col]
            return row

    def get_column(self, column, indexes=None, as_list=False):
        """Get row values from column.

        :param columns: Name for column
        :param indexes: Row indexes to include, or all if not given
        :param as_list: Return column as dictionary, instead of list
        """
        indexes = if_none(indexes, self.index)
        col = self.column_location(column)

        if as_list:
            column = []
            for index in indexes:
                idx = self.index_location(index)
                column.append(self._data[idx][col])
            return column
        else:
            column = {}
            for index in indexes:
                idx = self.index_location(index)
                column[idx] = self._data[idx][col]
            return column

    def get_table(self, indexes=None, columns=None, as_list=False):
        """Get a new table from all cells matching indexes and columns."""
        indexes = if_none(indexes, self.index)
        columns = if_none(columns, self._columns)

        if indexes == self.index and columns == self._columns:
            return self.copy()

        idxs = [self.index_location(index) for index in indexes]
        cols = [self.column_location(column) for column in columns]
        data = [[self._data[idx][col] for col in cols] for idx in idxs]

        if as_list:
            return data
        else:
            return Table(data=data, columns=columns)

    def get_slice(self, start: Optional[Index] = None, end: Optional[Index] = None):
        """Get a new table from rows between start and end index."""
        index = self._slice_index(slice(start, end))
        return self.get_table(index, self._columns)

    def _add_row(self, index):
        """Add a new empty row into the table."""
        if index is None:
            index = self.size

        if index < self.size:
            raise ValueError(f"Duplicate row index: {index}")

        for empty in range(self.size, index):
            self._add_row(empty)

        self._data.append([None] * len(self._columns))

        return self.size - 1

    def _add_column(self, column):
        """Add a new empty column into the table."""
        if column is None:
            column = len(self._columns)

        if column in self._columns:
            raise ValueError(f"Duplicate column name: {column}")

        if isinstance(column, int):
            assert column >= len(self._columns)
            for empty in range(len(self._columns), column):
                self._add_column(empty)

        self._columns.append(column)
        for idx in self.index:
            row = self._data[idx]
            row.append(None)

        return len(self._columns) - 1

    def set(self, indexes=None, columns=None, values=None):
        """Sets multiple cell values at a time.

        Both `indexes` and `columns` can be scalar or list-like,
        which enables setting individual cells, rows/columns, or regions.

        If `values` is scalar, all matching cells will be set to that value.
        Otherwise the length should match the cell count defined by the
        other parameters.
        """
        indexes = to_list(if_none(indexes, self.index))
        columns = to_list(if_none(columns, self._columns))

        size = len(indexes) + len(columns)
        values = to_list(values, size=size)
        if not len(values) == size:
            raise ValueError("Values size does not match indexes and columns")

        for index in indexes:
            idx = self.index_location(index)
            for column in columns:
                col = self.column_location(column)
                self.set_cell(index, column, values[idx + col])

    def set_cell(self, index, column, value):
        """Set individual cell value.
        If either index or column is missing, they are created.
        """
        try:
            idx = self.index_location(index)
        except (IndexError, ValueError):
            idx = self._add_row(index)

        try:
            col = self.column_location(column)
        except (IndexError, ValueError):
            col = self._add_column(column)

        self._data[idx][col] = value

    def set_row(self, index, values):
        """Set values in row. If index is missing, it is created."""
        try:
            idx = self.index_location(index)
        except (IndexError, ValueError):
            idx = self._add_row(index)

        column_values = self._column_value_getter(values)
        row = [column_values(values, column) for column in self._columns]

        self._data[idx] = row

    def set_column(self, column, values):
        """Set values in column. If column is missing, it is created."""
        values = to_list(values, size=self.size)

        if len(values) != self.size:
            raise ValueError(
                f"Values length ({len(values)}) should match data length ({self.size})"
            )

        if column not in self._columns:
            self._add_column(column)

        for index in self.index:
            self.set_cell(index, column, values[index])

    def append_row(self, row=None):
        """Append new row to table."""
        self.set_row(self.size, row)

    def append_rows(self, rows):
        """Append multiple rows to table."""
        for row in rows:
            self.append_row(row)

    def append_column(self, column=None, values=None):
        if column is not None and column in self._columns:
            raise ValueError(f"Column already exists: {column}")

        self.set_column(column, values)

    def delete_rows(self, indexes: Union[Index, List[Index]]):
        """Remove rows with matching indexes."""
        indexes = [self.index_location(idx) for idx in to_list(indexes)]

        unknown = set(indexes) - set(self.index)
        if unknown:
            names = ", ".join(str(name) for name in unknown)
            raise ValueError(f"Unable to remove unknown rows: {names}")

        for index in sorted(indexes, reverse=True):
            del self._data[index]

    def delete_columns(self, columns):
        """Remove columns with matching names."""
        columns = to_list(columns)

        unknown = set(columns) - set(self._columns)
        if unknown:
            names = ", ".join(str(name) for name in unknown)
            raise ValueError(f"Unable to remove unknown columns: {names}")

        for column in columns:
            col = self.column_location(column)
            for idx in self.index:
                del self._data[idx][col]
            del self._columns[col]

    def append_table(self, table):
        """Append data from table to current data."""
        if not table:
            return

        indexes = []
        for idx in table.index:
            index = self.size + idx
            indexes.append(index)

        self.set(indexes=indexes, columns=table.columns, values=table.data)

    def sort_by_column(self, columns, ascending=False):
        """Sort table by columns."""
        columns = to_list(columns)

        # Create sort criteria list, with each row as tuple of column values
        values = (self.get_column(column, as_list=True) for column in columns)
        values = list(zip(*values))
        assert len(values) == self.size

        def sorter(row):
            """Sort table by given values, while allowing for disparate types.
            Order priority:
                - Values by typename
                - Numeric types
                - None values
            """
            criteria = []
            for value in row[1]:  # Ignore enumeration
                criteria.append(
                    (
                        value is not None,
                        "" if isinstance(value, Number) else type(value).__name__,
                        value,
                    )
                )
            return criteria

        # Store original index order using enumerate() before sort,
        # and use it to sort data later
        values = sorted(enumerate(values), key=sorter, reverse=not ascending)
        idxs = [value[0] for value in values]

        # Re-order data
        self._data = [self._data[idx] for idx in idxs]

    def group_by_column(self, column):
        """Group rows by column value and return as list of tables."""
        ref = self.copy()
        ref.sort_by_column(column)

        col = self.column_location(column)
        groups = groupby(ref.data, itemgetter(col))

        result = []
        ref.clear()
        for _, group in groups:
            table = ref.copy()
            table.append_rows(group)
            result.append(table)

        return result

    def _filter(self, condition: RowCondition):
        filtered = list(filter(lambda idx: not condition(idx), self.index))
        self.delete_rows(filtered)

    def filter_all(self, condition: RowCondition):
        """Remove rows by evaluating `condition` for every row.

        The filtering will be done in-place and all the rows evaluating as falsy
        through the provided condition will be removed.
        """

        def _check_row(index: int) -> bool:
            row = self.get_row(index)
            return condition(row)

        self._filter(_check_row)

    def filter_by_column(self, column: Column, condition: CellCondition):
        """Remove rows by evaluating `condition` for cells in `column`.

        The filtering will be done in-place and all the rows where it evaluates to
        falsy are removed.
        """

        def _check_cell(index: int) -> bool:
            cell = self.get_cell(index, column)
            return condition(cell)

        self._filter(_check_cell)

    def iter_lists(self, with_index=True):
        """Iterate rows with values as lists."""
        for idx, row in zip(self.index, self._data):
            if with_index:
                yield idx, list(row)
            else:
                yield list(row)

    def iter_dicts(self, with_index=True) -> Generator[Dict[Column, Any], None, None]:
        """Iterate rows with values as dicts."""
        for index in self.index:
            row = {"index": index} if with_index else {}
            for column in self._columns:
                row[column] = self.get_cell(index, column)
            yield row

    def iter_tuples(self, with_index=True, name="Row"):
        """Iterate rows with values as namedtuples.
        Converts column names to valid Python identifiers,
        e.g. "First Name" -> "First_Name"
        """
        columns = {column: to_identifier(column) for column in self._columns}

        fields = ["index"] if with_index else []
        fields.extend(columns.values())

        container = namedtuple(name, fields)
        for row in self.iter_dicts(with_index):
            row = {columns[k]: v for k, v in row.items()}
            yield container(**row)

    def to_list(self, with_index=True):
        """Convert table to list representation."""
        export = []

        for index in self.index:
            row = OrderedDict()
            if with_index:
                row["index"] = index
            for column in self._columns:
                row[column] = self.get_cell(index, column)
            export.append(row)

        return export

    def to_dict(self, with_index=True):
        """Convert table to dict representation."""
        export = OrderedDict()

        if with_index:
            export["index"] = self.index

        for column in self._columns:
            export[column] = []

        for index in self.index:
            for column in self._columns:
                value = self.get_cell(index, column)
                export[column].append(value)

        return export


class Tables:
    """`Tables` is a library for manipulating tabular data inside Robot Framework.

    It can import data from various sources and apply different operations to it.
    Common use-cases are reading and writing CSV files, inspecting files in
    directories, or running tasks using existing Excel data.

    **Import types**

    The data a table can be created from can be of two main types:

    1. An iterable of individual rows, like a list of lists, or list of dictionaries
    2. A dictionary of columns, where each dictionary value is a list of values

    For instance, these two input values:

    .. code-block:: python

        data1 = [
            {"name": "Mark", "age": 58},
            {"name": "John", "age": 22},
            {"name": "Adam", "age": 67},
        ]

        data2 = {
            "name": ["Mark", "John", "Adam"],
            "age":  [    58,     22,     67],
        }

    Would both result in the following table:

    +-------+------+-----+
    | Index | Name | Age |
    +=======+======+=====+
    | 0     | Mark | 58  |
    +-------+------+-----+
    | 1     | John | 22  |
    +-------+------+-----+
    | 2     | Adam | 67  |
    +-------+------+-----+

    **Indexing columns and rows**

    Columns can be referred to in two ways: either with a unique string
    name or their position as an integer. Columns can be named either when
    the table is created, or they can be (re)named dynamically with keywords.
    The integer position can always be used, and it starts from zero.

    For instance, a table with columns "Name", "Age", and "Address" would
    allow referring to the "Age" column with either the name "Age" or the
    number 1.

    Rows do not have a name, but instead only have an integer index. This
    index also starts from zero. Keywords where rows are indexed also support
    negative values, which start counting backwards from the end.

    For instance, in a table with five rows, the first row could be referred
    to with the number 0. The last row could be accessed with either 4 or
    -1.

    **Examples**

    **Robot Framework**

    The `Tables` library can load tabular data from various other libraries
    and manipulate it inside Robot Framework.

    .. code-block:: robotframework

        *** Settings ***
        Library    RPA.Tables

        *** Keywords ***
        Files to Table
            ${files}=    List files in directory    ${CURDIR}
            ${files}=    Create table    ${files}
            Filter table by column    ${files}    size  >=  ${1024}
            FOR    ${file}    IN    @{files}
                Log    ${file}[name]
            END
            Write table to CSV    ${files}    ${OUTPUT_DIR}${/}files.csv

    **Python**

    The library is also available directly through Python, where it
    is easier to handle multiple different tables or do more bespoke
    manipulation operations.

    .. code-block:: python

        from RPA.Tables import Tables

        library = Tables()
        orders = library.read_table_from_csv(
            "orders.csv", columns=["name", "mail", "product"]
        )

        customers = library.group_table_by_column(rows, "mail")
        for customer in customers:
            for order in customer:
                add_cart(order)
            make_order()
    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    @staticmethod
    def _requires_table(obj: Any):
        if not isinstance(obj, Table):
            raise TypeError("Keyword requires Table object")

    def create_table(
        self, data: Data = None, trim: bool = False, columns: List[str] = None
    ) -> Table:
        """Create Table object from data.

        Data can be a combination of various iterable containers, e.g.
        list of lists, list of dicts, dict of lists.

        :param data:    Source data for table
        :param trim:    Remove all empty rows from the end of the worksheet,
                        default `False`
        :param columns: Names of columns (optional)
        :return:        Table object

        See the main library documentation for more information about
        supported data types.

        Example:

        .. code-block:: robotframework

            # Create a new table using a Dictionary of Lists
            # Because of the dictionary keys the column names will be automatically set
            @{Table_Data_name}=    Create List    Mark    John    Amy
            @{Table_Data_age}=    Create List    ${58}    ${22}    ${67}
            &{Table_Data}=    Create Dictionary
            ...    name=${Table_Data_name}
            ...    age=${Table_Data_age}
            ${table}=    Create Table    ${Table_Data}
        """
        table = Table(data, columns)

        if trim:
            self.trim_empty_rows(table)
            self.trim_column_names(table)

        self.logger.info("Created table: %s", table)
        notebook_table(self.table_head(table, 10))

        return table

    def export_table(
        self, table: Table, with_index: bool = False, as_list: bool = True
    ) -> Union[List, Dict]:
        """Convert a table object into standard Python containers.

        :param table:       Table to convert to dict
        :param with_index:  Include index in values
        :param as_list:     Export data as list instead of dict
        :return:            A List or Dictionary that represents the table

        Example:

        .. code-block:: robotframework

            ${orders}=       Read worksheet as table    orders.xlsx
            Sort table by column    ${orders}    CustomerId
            ${export}=       Export table    ${orders}
            # The following keyword expects a dictionary:
            Write as JSON    ${export}
        """
        self._requires_table(table)
        if as_list:
            return table.to_list(with_index)
        else:
            return table.to_dict(with_index)

    def copy_table(self, table: Table) -> Table:
        """Make a copy of a table object.

        :param table:   Table to copy
        :return:        Table object

        ${table_copy}=    Copy table    ${table}
        """
        self._requires_table(table)
        return table.copy()

    def clear_table(self, table: Table):
        """Clear table in-place, but keep columns.

        :param table:   Table to clear

        Example:

        .. code-block:: robotframework

            Clear table    ${table}
        """
        self._requires_table(table)
        table.clear()

    def merge_tables(self, *tables: Table, index: Optional[str] = None) -> Table:
        """Create a union of two tables and their contents.

        :param tables: Tables to merge
        :param index:  Column name to use as index for merge
        :return:       Table object

        By default rows from all tables are appended one after the other.
        Optionally a column name can be given with ``index``, which is
        used to merge rows together.

        Example:

        For instance, a ``name`` column could be used to identify
        unique rows and the merge operation should overwrite values
        instead of appending multiple copies of the same name.

        ====== =====
        Name   Price
        ====== =====
        Egg    10.0
        Cheese 15.0
        Ham    20.0
        ====== =====

        ====== =====
        Name   Stock
        ====== =====
        Egg    12.0
        Cheese 99.0
        Ham    0.0
        ====== =====

        .. code-block:: robotframework

            ${products}=    Merge tables    ${prices}    ${stock}    index=Name
            FOR    ${product}    IN    @{products}
                Log many
                ...    Product: ${product}[Name]
                ...    Price: ${product}[Price]
                ...    Stock: ${product}[Stock]
            END
        """
        if index is None:
            return self._merge_by_append(tables)
        else:
            return self._merge_by_index(tables, index)

    def _merge_by_append(self, tables: Tuple[Table, ...]):
        """Merge tables by appending columns and rows."""
        columns = uniq(column for table in tables for column in table.columns)

        merged = Table(columns=columns)
        for table in tables:
            merged.append_rows(table)

        return merged

    def _merge_by_index(self, tables: Tuple[Table, ...], index: str):
        """Merge tables by using a column as shared key."""
        columns = uniq(column for table in tables for column in table.columns)
        merged = Table(columns=columns)

        seen = {}

        def find_index(row):
            """Find index for row, if key already exists."""
            value = row[index]
            if value in seen:
                return seen[value]
            for row_ in merged.iter_dicts(True):
                if row_[index] == value:
                    seen[value] = row_["index"]
                    return row_["index"]
            return None

        for table in tables:
            for row in table.iter_dicts(False):
                row_index = find_index(row)
                if row_index is None:
                    merged.append_row(row)
                else:
                    for column, value in row.items():
                        merged.set_cell(row_index, column, value)

        return merged

    def get_table_dimensions(self, table: Table) -> Tuple[int, int]:
        """Return table dimensions, as (rows, columns).

        :param table:    Table to inspect
        :return:         Two integer values that represent the number
                         of rows and columns

        Example:

        .. code-block:: robotframework

            ${rows}  ${columns}=    Get table dimensions    ${table}
            Log    Table has ${rows} rows and ${columns} columns.
        """
        self._requires_table(table)
        notebook_print(text=table.dimensions)
        return table.dimensions

    def rename_table_columns(
        self, table: Table, names: List[Union[str, None]], strict: bool = False
    ):
        """Renames columns in the Table with given values. Columns with
        name as ``None`` will use the previous value.

        :param table:   Table to modify
        :param names:   List of new column names
        :param strict:  If True, raises ValueError if column lengths
                        do not match

        The renaming will be done in-place.

        Examples:

        .. code-block:: robotframework

            # Initially set the column names
            ${columns}=    Create list   First  Second  Third
            Rename table columns    ${table}    ${columns}
            # First, Second, Third

            # Update the first and second column names to Uno and Dos
            ${columns}=    Create list   Uno  Dos
            Rename table columns    ${table}    ${columns}
            # Uno, Dos, Third
        """
        self._requires_table(table)
        before = table.columns

        if strict and len(before) != len(names):
            raise ValueError("Column lengths do not match")

        after = []
        for old, new in zip_longest(before, names):
            if old is None:
                break
            elif new is None:
                after.append(old)
            else:
                after.append(new)

        table.columns = after

    def add_table_column(
        self, table: Table, name: Optional[str] = None, values: Any = None
    ):
        """Append a column to a table.

        :param table:   Table to modify
        :param name:    Name of new column
        :param values:  Value(s) for new column

        The ``values`` can either be a list of values, one for each row, or
        one single value that is set for all rows.

        Examples:

        .. code-block:: robotframework

            # Add empty column
            Add table column    ${table}

            # Add empty column with name
            Add table column    ${table}    name=Home Address

            # Add new column where every every row has the same value
            Add table column    ${table}    name=TOS    values=${FALSE}

            # Add new column where every row has a unique value
            ${is_first}=    Create list    ${TRUE}    ${FALSE}    ${FALSE}
            Add table column    ${table}    name=IsFirst    values=${is_first}
        """
        self._requires_table(table)
        table.append_column(name, values)

    def add_table_row(self, table: Table, values: Any = None):
        """Append rows to a table.

        :param table:   Table to modify
        :param values:  Value(s) for new row

        The ``values`` can either be a list of values, or a dictionary
        where the keys match current column names. Values for unknown
        keys are discarded.

        It can also be a single value that is set for all columns,
        which is ``None`` by default.

        Examples:

        .. code-block:: robotframework

            # Add empty row
            Add table row    ${table}

            # Add row where every column has the same value
            Add table row    ${table}    Unknown

            # Add values per column
            ${values}=    Create dictionary    Username=Mark    Mail=mark@robocorp.com
            Add table row    ${table}    ${values}
        """
        self._requires_table(table)
        table.append_row(values)

    def get_table_row(
        self, table: Table, row: Index, as_list: bool = False
    ) -> Union[Dict, List]:
        """Get a single row from a table.

        :param table:   Table to read
        :param row:     Row to read
        :param as_list: Return list instead of dictionary
        :return:        Dictionary or List of table row

        Examples:

        .. code-block:: robotframework

            # returns the first row in the table
            ${first}=    Get table row    ${orders}

            # returns the last row in the table
            ${last}=      Get table row    ${orders}    -1    as_list=${TRUE}
        """
        self._requires_table(table)
        values = table.get_row(row, as_list=as_list)
        notebook_print(text=values)
        return values

    def get_table_column(self, table: Table, column: Column) -> List:
        """Get all values for a single column in a table.

        :param table:   Table to read
        :param column:  Column to read
        :return:        List of the rows in the selected column

        Example:

        .. code-block:: robotframework

            ${emails}=    Get table column    ${users}    E-Mail Address
        """
        self._requires_table(table)
        col = table.get_column(column, as_list=True)
        return col

    def set_table_row(self, table: Table, row: Index, values: Any):
        """Assign values to a row in the table.

        :param table:   Table to modify
        :param row:     Row to modify
        :param values:  Value(s) to set

        The ``values`` can either be a list of values, or a dictionary
        where the keys match current column names. Values for unknown
        keys are discarded.

        It can also be a single value that is set for all columns.

        Examples:

        .. code-block:: robotframework

            ${columns}=  Create list     One  Two  Three
            ${table}=    Create table    columns=${columns}

            ${values}=   Create list     1  2  3
            Set table row    ${table}    0    ${values}

            ${values}=   Create dictionary    One=1  Two=2  Three=3
            Set table row    ${table}    1    ${values}

            Set table row    ${table}    2    ${NONE}
        """
        self._requires_table(table)
        table.set_row(row, values)

    def set_table_column(self, table: Table, column: Column, values: Any):
        """Assign values to a column in the table.

        :param table:   Table to modify
        :param column:  Column to modify
        :param values:  Value(s) to set

        The ``values`` can either be a list of values, one for each row, or
        one single value that is set for all rows.

        Examples:

        .. code-block:: robotframework

            # Set different value for each row (sizes must match)
            ${ids}=    Create list    1  2  3  4  5
            Set table column    ${users}    userId    ${ids}

            # Set the same value for all rows
            Set table column    ${users}    email     ${NONE}
        """
        self._requires_table(table)
        table.set_column(column, values)

    def pop_table_row(
        self, table: Table, row: Optional[Index] = None, as_list: bool = False
    ) -> Union[Dict, List]:
        """Remove row from table and return it.

        :param table:   Table to modify
        :param row:     Row index, pops first row if none given
        :param as_list: Return list instead of dictionary
        :return:        Dictionary or List of the removed, popped, row

        Examples:

        .. code-block:: robotframework

            # Pop the firt row in the table and discard it
            Pop table row    ${orders}

            # Pop the last row in the table and store it
            ${row}=      Pop table row    ${data}    -1    as_list=${TRUE}
        """
        self._requires_table(table)
        row = if_none(row, table.index[0])

        values = table.get_row(row, as_list=as_list)
        table.delete_rows(row)
        return values

    def pop_table_column(
        self, table: Table, column: Optional[Column] = None
    ) -> Union[Dict, List]:
        """Remove column from table and return it.

        :param table:   Table to modify
        :param column:  Column to remove
        :return:        Dictionary or List of the removed, popped, column

        Examples:

        .. code-block:: robotframework

            # Remove column from table and discard it
            Pop table column    ${users}   userId

            # Remove column from table and iterate over it
            ${ids}=    Pop table column    ${users}    userId
            FOR    ${id}    IN    @{ids}
                Log    User id: ${id}
            END
        """
        self._requires_table(table)
        column: Column = if_none(column, table.columns[0])

        values = self.get_table_column(table, column)
        table.delete_columns(column)
        return values

    def get_table_slice(
        self, table: Table, start: Optional[Index] = None, end: Optional[Index] = None
    ) -> Union[Table, List[List]]:
        """Return a new Table from a range of given Table rows.

        :param table:   Table to read from
        :param start:   Start index (inclusive)
        :param start:   End index (exclusive)
        :return:        Table object of the selected rows

        If ``start`` is not defined, starts from the first row.
        If ``end`` is not defined, stops at the last row.

        Examples:

        .. code-block:: robotframework

            # Get all rows except first five
            ${slice}=    Get table slice    ${table}    start=5

            # Get rows at indexes 5, 6, 7, 8, and 9
            ${slice}=    Get table slice    ${table}    start=5    end=10

            # Get all rows except last five
            ${slice}=    Get table slice    ${table}    end=-5
        """
        self._requires_table(table)
        return table.get_slice(start, end)

    def set_row_as_column_names(self, table: Table, row: Index):
        """Set existing row as names for columns.

        :param table: Table to modify
        :param row:   Row to use as column names

        Example:

        .. code-block:: robotframework

            # Set the column names based on the first row
            Set row as column names    ${table}    0
        """
        values = self.pop_table_row(table, row, as_list=True)
        table.columns = values

    def table_head(
        self, table: Table, count: int = 5, as_list: bool = False
    ) -> Union[Table, List[List]]:
        """Return first ``count`` rows from a table.

        :param table:   Table to read from
        :param count:   Number of lines to read
        :param as_list: Return list instead of Table
        :return:        Return Table object or List of the selected rows

        Example:

        .. code-block:: robotframework

            # Get the first 10 employees
            ${first}=    Table head    ${employees}    10
        """
        self._requires_table(table)
        return table.head(count, as_list)

    def table_tail(
        self, table: Table, count: int = 5, as_list: bool = False
    ) -> Union[Table, List[List]]:
        """Return last ``count`` rows from a table.

        :param table:   Table to read from
        :param count:   Number of lines to read
        :param as_list: Return list instead of Table
        :return:        Return Table object or List of the selected rows

        Example:

        .. code-block:: robotframework

            # Get the last 10 orders
            ${latest}=    Table tail    ${orders}    10
        """
        self._requires_table(table)
        return table.tail(count, as_list)

    def get_table_cell(self, table: Table, row: Index, column: Column) -> Any:
        """Get a cell value from a table.

        :param table:   Table to read from
        :param row:     Row of cell
        :param column:  Column of cell
        :return:        Cell value

        Examples:

        .. code-block:: robotframework

            # Get the value in the first row and first column
            Get table cell    ${table}    0    0

            # Get the value in the last row and first column
            Get table cell    ${table}   -1    0

            # Get the value in the last row and last column
            Get table cell    ${table}   -1    -1

            # Get the value in the third row and column "Name"
            Get table cell    ${table}    2    Name
        """
        self._requires_table(table)
        return table.get_cell(row, column)

    def set_table_cell(self, table: Table, row: Index, column: Column, value: Any):
        """Set a cell value in a table.

        :param table:   Table to modify to
        :param row:     Row of cell
        :param column:  Column of cell
        :param value:   Value to set

        Examples:

        .. code-block:: robotframework

            # Set the value in the first row and first column to "First"
            Set table cell    ${table}    0    0       First

            # Set the value in the last row and first column to "Last"
            Set table cell    ${table}   -1    0       Last

            # Set the value in the last row and last column to "Corner"
            Set table cell    ${table}   -1    -1       Corner

            # Set the value in the third row and column "Name" to "Unknown"
            Set table cell    ${table}    2    Name    Unknown
        """
        self._requires_table(table)
        table.set_cell(row, column, value)

    def find_table_rows(self, table: Table, column: Column, operator: str, value: Any):
        """Find all the rows in a table which match a condition for a given column.

        :param table: Table to search into.
        :param column: Name or position of the column to compare with.
        :param operator: Comparison operator used with every cell value on the
            specified column.
        :param value: Value to compare against.
        :return: New `Table` object containing all the rows matching the condition.

        Supported operators:

        ============ ========================================
        Operator     Description
        ============ ========================================
        >            Cell value is larger than
        <            Cell value is smaller than
        >=           Cell value is larger or equal than
        <=           Cell value is smaller or equal than
        ==           Cell value is equal to
        !=           Cell value is not equal to
        is           Cell value is the same object
        not is       Cell value is not the same object
        contains     Cell value contains given value
        not contains Cell value does not contain given value
        in           Cell value is in given value
        not in       Cell value is not in given value
        ============ ========================================

        Returns the matches as a new `Table` instance.

        Examples:

        .. code-block:: robotframework

            # Find all rows where price is over 200
            @{rows} =    Find table rows    ${table}    Price  >  ${200}

            # Find all rows where the status does not contain "removed"
            @{rows} =    Find table rows    ${table}    Status  not contains  removed
        """
        self._requires_table(table)

        condition = to_condition(operator, value)

        matches = []
        for index in table.index:
            cell = table.get_cell(index, column)
            if condition(cell):
                matches.append(index)

        return table.get_table(matches)

    def sort_table_by_column(
        self, table: Table, column: Column, ascending: bool = True
    ):
        """Sort a table in-place according to ``column``.

        :param table:       Table to sort
        :param column:      Column to sort with
        :param ascending:   Table sort order

        Examples:

        .. code-block:: robotframework

            # Sorts the `order_date` column ascending
            Sort table by column    ${orders}    order_date

            # Sorts the `order_date` column descending
            Sort table by column    ${orders}    order_date    ascending=${FALSE}
        """
        self._requires_table(table)
        table.sort_by_column(column, ascending=ascending)

    def group_table_by_column(self, table: Table, column: Column) -> List[Table]:
        """Group a table by ``column`` and return a list of grouped Tables.

        :param table:   Table to use for grouping
        :param column:  Column which is used as grouping criteria
        :return:        List of Table objects

        Example:

        .. code-block:: robotframework

            # Groups rows of matching customers from the `customer` column
            # and returns the groups or rows as Tables
            @{groups}=    Group table by column    ${orders}    customer
            # An example of how to use the List of Tables once returned
            FOR    ${group}    IN    @{groups}
                # Process all orders for the customer at once
                Process order    ${group}
            END
        """
        self._requires_table(table)
        groups = table.group_by_column(column)
        self.logger.info("Found %s groups", len(groups))
        return groups

    def filter_table_by_column(
        self, table: Table, column: Column, operator: str, value: Any
    ):
        """Remove all rows where column values don't match the
        given condition.

        :param table:     Table to filter
        :param column:    Column to filter with
        :param operator:  Filtering operator, e.g. >, <, ==, contains
        :param value:     Value to compare column to (using operator)

        See the keyword ``Find table rows`` for all supported operators
        and their descriptions.

        The filtering will be done in-place.

        Examples:

        .. code-block:: robotframework

            # Only accept prices that are non-zero
            Filter table by column    ${table}   price  !=  ${0}

            # Remove uwnanted product types
            @{types}=    Create list    Unknown    Removed
            Filter table by column    ${table}   product_type  not in  ${types}
        """
        self._requires_table(table)

        condition = to_condition(operator, value)

        before = len(table)
        table.filter_by_column(column, condition)
        after = len(table)

        self.logger.info("Filtered %d rows", after - before)

    def filter_table_with_keyword(self, table: Table, name: str, *args):
        """Run a keyword for each row of a table, then remove all rows where the called
        keyword returns a falsy value.

        Can be used to create custom RF keyword based filters.

        :param table: Table to modify.
        :param name: Keyword name used as filter.
        :param args: Additional keyword arguments to be passed. (optional)

        The row object will be given as the first argument to the filtering keyword.
        """
        self._requires_table(table)

        def condition(row: Row) -> bool:
            return BuiltIn().run_keyword(name, row, *args)

        before = len(table)
        table.filter_all(condition)
        after = len(table)

        self.logger.info("Removed %d row(s)", before - after)

    def map_column_values(self, table: Table, column: Column, name: str, *args):
        """Run a keyword for each cell in a given column, and replace its content with
        the return value.

        Can be used to easily convert column types or values in-place.

        :param table: Table to modify.
        :param column: Column to modify.
        :param name: Mapping keyword name.
        :param args: Additional keyword arguments. (optional)

        The cell value will be given as the first argument to the mapping keyword.

        Examples:

        .. code-block:: robotframework

            # Convert all columns values to a different type
            Map column values    ${table}    Price    Convert to integer

            # Look up values with a custom keyword
            Map column values    ${table}    User     Map user ID to name
        """
        self._requires_table(table)

        values = []
        for index in table.index:
            cell = table.get_cell(index, column)
            output = BuiltIn().run_keyword(name, cell, *args)
            values.append(output)

        table.set_column(column, values)

    def filter_empty_rows(self, table: Table):
        """Remove all rows from a table which have only ``None`` values.

        :param table:   Table to filter

        The filtering will be done in-place.

        Example:

        .. code-block:: robotframework

            Filter empty rows    ${table}
        """
        self._requires_table(table)

        empty = []
        for idx, row in table.iter_lists():
            if all(value is None for value in row):
                empty.append(idx)

        table.delete_rows(empty)

    def trim_empty_rows(self, table: Table):
        """Remove all rows from the *end* of a table
        which have only ``None`` as values.

        :param table:    Table to filter

        The filtering will be done in-place.

        Example:

        .. code-block:: robotframework

            Trim empty rows    ${table}
        """
        self._requires_table(table)

        empty = []
        for idx in reversed(table.index):
            row = table[idx]
            if any(value is not None for value in row):
                break
            empty.append(idx)

        table.delete_rows(empty)

    def trim_column_names(self, table: Table):
        """Remove all extraneous whitespace from column names.

        :param table:    Table to filter

        The filtering will be done in-place.

        Example:

        .. code-block:: robotframework

            # This example will take colums such as:
            # "One", "Two ", "  Three "
            # and trim them to become the below:
            # "One", "Two", "Three"
            Trim column names     ${table}
        """
        self._requires_table(table)
        table.columns = [
            column.strip() if isinstance(column, str) else column
            for column in table.columns
        ]

    @keyword("Read table from CSV")
    def read_table_from_csv(
        self,
        path: str,
        header: Optional[bool] = None,
        columns: Optional[List[str]] = None,
        dialect: Optional[Dialect] = None,
        delimiters: Optional[str] = None,
        column_unknown: str = "Unknown",
        encoding: Optional[str] = None,
    ) -> Table:
        """Read a CSV file as a table.

        :param path:            Path to CSV file
        :param header:          CSV file includes header
        :param columns:         Names of columns in resulting table
        :param dialect:         Format of CSV file
        :param delimiters:      String of possible delimiters
        :param column_unknown:  Column name for unknown fields
        :param encoding:        Text encoding for input file,
                                uses system encoding by default
        :return:                Table object

        By default attempts to deduce the CSV format and headers
        from a sample of the input file. If it's unable to determine
        the format automatically, the dialect and header will
        have to be defined manually.

        Valid ``dialect`` values are ``excel``, ``excel-tab``, and ``unix``,
        and ``header`` is boolean argument (``True``/``False``). Optionally a
        set of valid ``delimiters`` can be given as a string.

        The ``columns`` argument can be used to override the names of columns
        in the resulting table. The amount of columns must match the input
        data.

        If the source data has a header and rows have more fields than
        the header defines, the remaining values are put into the column
        given by ``column_unknown``. By default it has the value "Unknown".

        Examples:

        .. code-block:: robotframework

            # Source dialect is deduced automatically
            ${table}=    Read table from CSV    export.csv
            Log   Found columns: ${table.columns}

            # Source dialect is known and given explicitly
            ${table}=    Read table from CSV    export-excel.csv    dialect=excel
            Log   Found columns: ${table.columns}
        """
        sniffer = csv.Sniffer()
        with open(path, newline="", encoding=encoding) as fd:
            sample = fd.readline()

        if dialect is None:
            dialect_name = sniffer.sniff(sample, delimiters)
        elif isinstance(dialect, Dialect):
            dialect_name = dialect.value
        else:
            dialect_name = Dialect(dialect).value

        if header is None:
            header = sniffer.has_header(sample)

        with open(path, newline="", encoding=encoding) as fd:
            if header:
                reader = csv.DictReader(
                    fd, dialect=dialect_name, restkey=str(column_unknown)
                )
            else:
                reader = csv.reader(fd, dialect=dialect_name)
            rows = list(reader)

        table = Table(rows, columns)
        notebook_table(self.table_head(table, 10))

        if header and column_unknown in table.columns:
            self.logger.warning(
                "CSV file (%s) had fields not defined in header, "
                "which can be the result of a wrong dialect",
                path,
            )

        return table

    @keyword("Write table to CSV")
    def write_table_to_csv(
        self,
        table: Table,
        path: str,
        header: bool = True,
        dialect: Dialect = Dialect.Excel,
        encoding: Optional[str] = None,
        delimiter: Optional[str] = ",",
    ):
        """Write a table as a CSV file.

        :param table:    Table to write
        :param path:     Path to write to
        :param header:   Write columns as header to CSV file
        :param dialect:  The format of output CSV
        :param encoding: Text encoding for output file,
                         uses system encoding by default
        :param delimiter: Delimiter character between columns

        Valid ``dialect`` values are ``Excel``, ``ExcelTab``, and ``Unix``.

        Example:

        .. code-block:: robotframework

            ${sheet}=    Read worksheet as table    orders.xlsx    header=${TRUE}
            Write table to CSV    ${sheet}    output.csv
        """
        self._requires_table(table)

        if isinstance(dialect, str):
            dialect = Dialect(dialect)

        with open(path, mode="w", newline="", encoding=encoding) as fd:
            writer = csv.DictWriter(
                fd, fieldnames=table.columns, dialect=dialect.value, delimiter=delimiter
            )

            if header:
                writer.writeheader()

            for row in table.iter_dicts(with_index=False):
                writer.writerow(row)
